from sklearn.model_selection import ShuffleSplit
from sklearn.model_selection import StratifiedKFold
from sklearn.utils import check_random_state

from csrank.constants import DYAD_RANKING
from ..sushi_dataset_reader import SushiDatasetReader


class SushiDyadRankingDatasetReader(SushiDatasetReader):
    def __init__(self, random_state=None, **kwargs):
        super(SushiDyadRankingDatasetReader, self).__init__(
            learning_problem=DYAD_RANKING, **kwargs
        )
        self.random_state = check_random_state(random_state)

    def splitter(self, iter):
        for train_idx, test_idx in iter:
            yield self.X[train_idx], self.Xc[train_idx], self.Y[train_idx], self.X[
                test_idx
            ], self.Xc[train_idx], self.Y[test_idx]

    def get_train_test_datasets(self, n_datasets=5):
        cv_iter = StratifiedKFold(
            n_splits=n_datasets, shuffle=True, random_state=self.random_state
        )
        splits = list(cv_iter.split(self.X))
        return self.splitter(splits)

    def get_dataset_dictionaries(self):
        return self.Xc, self.X, self.Y

    def get_single_train_test_split(self):
        cv_iter = ShuffleSplit(
            n_splits=1, random_state=self.random_state, test_size=0.30
        )
        splits = list(cv_iter.split(self.X))
        return list(self.splitter(splits))[0]



import logging
import os

import numpy as np
from sklearn.utils import check_random_state

from csrank.constants import DISCRETE_CHOICE
from .util import sub_sampling_discrete_choices_from_relevance
from ..expedia_dataset_reader import ExpediaDatasetReader

logger = logging.getLogger(__name__)


class ExpediaDiscreteChoiceDatasetReader(ExpediaDatasetReader):
    def __init__(self, random_state=None, n_objects=5, **kwargs):
        super(ExpediaDiscreteChoiceDatasetReader, self).__init__(
            learning_problem=DISCRETE_CHOICE, **kwargs
        )
        self.random_state = check_random_state(random_state)
        self.n_objects = n_objects
        self.hdf5file_path = os.path.join(self.dirname, "exp_dc.h5")
        self.__load_dataset__()

    def sub_sampling_function(self, X, Y):
        return sub_sampling_discrete_choices_from_relevance(
            Xt=X, Yt=Y, n_objects=self.n_objects
        )

    def get_train_test_datasets(self, n_datasets):
        pass

    def _parse_dataset(self):
        self._filter_dataset()
        X = []
        Y = []
        clicks = []
        for name, group in self.train_df.groupby("srch_id"):
            del group["srch_id"]
            vals = group.values
            y = vals[:, -1] + vals[:, -2]
            if (2 in y) or (y.sum() == 1.0):
                X.append(vals[:, 0:-2])
                y[y == np.max(y)] = 1
                y[y != np.max(y)] = 0
                Y.append(y)
            else:
                clicks.append(y.sum())
        logger.info("Percentage of clicked {}".format(len(clicks) / len(Y)))
        return X, Y



"""An implementation of the scoring module for FATE estimators."""

import functools

import torch.nn as nn

from csrank.modules.collection import PairwiseEvaluation
from csrank.modules.instance_reduction import MeanAggregatedUtility
from csrank.modules.object_mapping import DenseNeuralNetwork


class FETAScoring(nn.Module):
    """Map instances to scores with the FETA approach.

    >>> from csrank.modules.object_mapping import DeterministicSumming
    >>> scoring = FETAScoring(
    ...     n_features=2,
    ...     pairwise_utility_module=DeterministicSumming,
    ... )

    Now let's define some problem instances.

    >>> object_a = [0.5, 0.8]
    >>> object_b = [1.5, 1.8]
    >>> object_c = [2.5, 2.8]
    >>> object_d = [3.5, 3.6]
    >>> object_e = [4.5, 4.6]
    >>> object_f = [5.5, 5.6]
    >>> # instance = list of objects to rank
    >>> instance_a = [object_a, object_b, object_c]
    >>> instance_b = [object_d, object_e, object_f]
    >>> import torch
    >>> instances = torch.tensor([instance_a, instance_b])

    >>> scoring(instances)
    tensor([[ 5.6000,  6.6000,  7.6000],
            [17.2000, 18.2000, 19.2000]])

    Parameters
    ----------
    n_features: int
        The number of features each object has.

    pairwise_utility_module: pytorch module with one integer parameter
        The module that should be used for pairwise utility estimations. Uses a
        simple linear mapping if not specified. You likely want to replace this
        with something more expressive such as a ``DenseNeuralNetwork``. This
        should take the size of the input values as its only parameter. You can
        use ``functools.partial`` if necessary. This corresponds to :math:`U`
        [1]_.

    zeroth_order_module: pytorch module with one integer parameter
        The module that should be used to evaluate objects in isolation. May be
        ``None``, in which case no zeroth order module is used. That is the
        default behavior. You may want to replace this with something like a
        ``DenseNeuralNetwork``. This should take the size of the input values
        as its only parameter. You can use ``functools.partial`` if necessary.
        This corresponds to :math:`U` This corresponds to :math:`U_0` [1]_.
    """

    def __init__(
        self,
        n_features,
        pairwise_utility_module=functools.partial(
            DenseNeuralNetwork, hidden_layers=3, units_per_hidden=20
        ),
        zeroth_order_module=None,
    ):
        super().__init__()
        # Compute the utility of each object in the context of every other
        # object. Use mean aggregation to come to a single utility value per
        # object. Ignore self-comparisons. Including those would be similar to
        # adding a zeroth-order model, but with a different scale and learned
        # included in the same utility network. That could make it harder to
        # learn.
        self.mean_aggregated_utilty = MeanAggregatedUtility(
            PairwiseEvaluation(pairwise_utility_module(2 * n_features)),
            exclude_self_comparison=True,
        )
        self.zeroth_order_module = (
            None if zeroth_order_module is None else zeroth_order_module(n_features)
        )

    def forward(self, instances, **kwargs):
        total_utility = self.mean_aggregated_utilty(instances)
        if self.zeroth_order_module is not None:
            # The result of the zeroth order utility module has a singleton
            # dimension, while the aggregated utility does not.
            zeroth_order_utility = self.zeroth_order_module(instances).reshape(
                total_utility.shape
            )
            total_utility += zeroth_order_utility
        return total_utility



"""An implementation of the scoring module for FATE estimators."""

import functools

import torch
import torch.nn as nn

from csrank.modules.instance_reduction import DeepSet
from csrank.modules.object_mapping import DenseNeuralNetwork


class FATEScoring(nn.Module):
    r"""Map instances to scores with the FATE approach.

    Let's show the FATE approach on an example. To simplify things, we'll use a
    simply identity-embedding. The FATE module will then aggregate the context
    by simply taking the average of the objects (feature-wise). To further
    simplify things the actual pairwise utility is just computed by the sum of
    all features of the object and the context.

    >>> import torch.nn as nn
    >>> from csrank.modules.object_mapping import DeterministicSumming
    >>> scoring = FATEScoring(
    ...     n_features=2,
    ...     pairwise_utility_module=DeterministicSumming,
    ...     embedding_module=nn.Identity,
    ... )

    Now let's define some problem instances.

    >>> object_a = [0.5, 0.8]
    >>> object_b = [1.5, 1.8]
    >>> object_c = [2.5, 2.8]
    >>> object_d = [3.5, 3.6]
    >>> object_e = [4.5, 4.6]
    >>> object_f = [5.5, 5.6]
    >>> # instance = list of objects to rank
    >>> instance_a = [object_a, object_b, object_c]
    >>> instance_b = [object_d, object_e, object_f]
    >>> import torch
    >>> instances = torch.tensor([instance_a, instance_b])

    Let's focus on the first instance in this example. The aggregated identity
    embedding is

    >>> embedding_1 = (object_a[0] + object_b[0] + object_c[0]) / 3
    >>> embedding_2 = (object_a[1] + object_b[1] + object_c[1]) / 3
    >>> (embedding_1, embedding_2)
    (1.5, 1.8)

    for the first and second feature respectively. So the utility of object_a
    within the context (defined by the mock sum utility) should be

    >>> embedding_1 + embedding_2 + object_a[0] + object_a[1]
    4.6

    Let's verify this:

    >>> scoring(instances)
    tensor([[ 4.6000,  6.6000,  8.6000],
            [16.2000, 18.2000, 20.2000]])

    As you can see, the scoring comes to the same result for the first object
    of the first instance.

    Parameters
    ----------
    n_features: int
        The number of features each object has.
    embedding_size: int
        The size of the embeddings that should be generated. Defaults to
        ``n_features`` if not specified.
    pairwise_utility_module: pytorch module with one integer parameter
        The module that should be used for pairwise utility estimations. Uses a
        simple linear mapping not specified. You likely want to replace this
        with something more expressive such as a ``DenseNeuralNetwork``. This
        should take the size of the input values as its only parameter. You can
        use ``functools.partial`` if necessary. This corresponds to
        :math:`U` in Figure 2 of [1]_.
    embedding_module: pytorch module with one integer parameter
        The module that should be used for the object embeddings. Its
        constructor should take two parameters: The size of the input and the
        size of the output. This corresponds to :math:`\Phi` in Figure 2 of
        [1]_. The default is a ``DenseNeuralNetwork`` with 5 hidden layers and
        64 units per hidden layer.

    References
    ----------
    .. [1] Pfannschmidt, K., Gupta, P., & Hüllermeier, E. (2019). Learning
    choice functions: Concepts and architectures. arXiv preprint
    arXiv:1901.10860.
    """

    def __init__(
        self,
        n_features,
        embedding_size=None,
        pairwise_utility_module=None,
        embedding_module=None,
    ):
        super().__init__()
        if embedding_size is None:
            embedding_size = n_features
        if pairwise_utility_module is None:
            pairwise_utility_module = functools.partial(
                nn.Linear,
                out_features=1,
            )
        if embedding_module is None:
            embedding_module = functools.partial(
                DenseNeuralNetwork, hidden_layers=5, units_per_hidden=64
            )

        self.embedding = DeepSet(
            n_features,
            embedding_size,
            embedding_module=embedding_module,
        )
        self.pairwise_utility_module = pairwise_utility_module(
            n_features + embedding_size
        )

    def forward(self, instances, **kwargs):
        n_objects = instances.size(1)
        contexts = self.embedding(instances)
        # Repeat each context for each object within the instance; This is then
        # a flat list of contexts. Then reshape to have a list of contexts per
        # instance.
        context_per_object = contexts.repeat_interleave(n_objects, dim=0).reshape_as(
            instances
        )
        pairs = torch.stack((instances, context_per_object), dim=-1)
        utilities = self.pairwise_utility_module(pairs.flatten(start_dim=-2)).squeeze()
        return utilities



from abc import ABCMeta

from csrank.constants import DISCRETE_CHOICE
from csrank.discrete_choice_losses import CategoricalHingeLossMax
from csrank.learner import SkorchInstanceEstimator
from csrank.util import convert_to_label_encoding

__all__ = ["DiscreteObjectChooser", "SkorchDiscreteChoiceFunction"]


class DiscreteObjectChooser(metaclass=ABCMeta):
    @property
    def learning_problem(self):
        return DISCRETE_CHOICE

    def predict_for_scores(self, scores):
        """
        Binary discrete choice vector :math:`y` represents the choices amongst the objects in :math:`Q`, such that
        :math:`y(k) = 1` represents that the object :math:`x_k` is chosen and :math:`y(k) = 0` represents
        it is not chosen. For choice to be discrete :math:`\\sum_{x_i \\in Q} y(i) = 1`. Predict discrete choices for
        the scores for a given collection of sets of objects (query sets).

        Parameters
        ----------
        scores : dict or numpy array
            Dictionary with a mapping from query set size to numpy arrays
            or a single numpy array of size containing scores of each object of size:
            (n_instances, n_objects)


        Returns
        -------
        Y : dict or numpy array
            Dictionary with a mapping from query set size to numpy arrays
            or a single numpy array containing predicted discrete choice vectors of size:
            (n_instances, n_objects)
        """
        if isinstance(scores, dict):
            result = dict()
            for n, s in scores.items():
                result[n] = s.argmax(axis=1)
                result[n] = convert_to_label_encoding(result[n], n)

        else:
            n = scores.shape[-1]
            result = scores.argmax(axis=1)
            result = convert_to_label_encoding(result, n)
        return result


class SkorchDiscreteChoiceFunction(DiscreteObjectChooser, SkorchInstanceEstimator):
    """Base estimator for torch-based discrete choice.

    This makes it very simple to derive new estimators with any given scoring
    module. Refer to skorch's documentation for supported parameters. For
    example the optimizer or the optimizer's learning rate could be overridden.

    Parameters
    ----------
    module : torch module (class)
        This is the scoring module. It should be an uninstantiated
        ``torch.nn.Module`` class that expects the number of features per
        object as its only parameter on initialization.

    criterion : torch criterion (class)
        The criterion that is used to evaluate and optimize the module.

    choice_size : int
        The size of the target choice set.

    **kwargs : skorch NeuralNet arguments
        All keyword arguments are passed to the constructor of
        ``skorch.NeuralNet``. See the documentation of that class for more
        details.
    """

    def __init__(
        self, module, criterion=CategoricalHingeLossMax, choice_size=1, **kwargs
    ):
        super().__init__(module=module, criterion=criterion, **kwargs)
        self.choice_size = choice_size



from itertools import product
import logging

import numpy as np

from csrank.learner import Learner
import csrank.numpy_util as npu
import csrank.theano_util as ttu
from csrank.util import print_dictionary
from .discrete_choice import DiscreteObjectChooser
from .likelihoods import create_weight_dictionary
from .likelihoods import fit_pymc3_model
from .likelihoods import likelihood_dict
from .likelihoods import LogLikelihood

try:
    import pymc3 as pm
    from pymc3.variational.callbacks import CheckParametersConvergence
except ImportError:
    from csrank.util import MissingExtraError

    raise MissingExtraError("pymc3", "probabilistic")

try:
    import theano
    from theano import tensor as tt
except ImportError:
    from csrank.util import MissingExtraError

    raise MissingExtraError("theano", "probabilistic")

logger = logging.getLogger(__name__)


class MixedLogitModel(DiscreteObjectChooser, Learner):
    def __init__(self, n_mixtures=4, loss_function="", regularization="l2", **kwargs):
        """
        Create an instance of the Mixed Logit model for learning the discrete choice function. In this model we
        assume weights of this model to be random due to which this model can learn different variations in choices
        amongst the individuals. The utility score for each object in query set :math:`Q` is defined as
        :math:`U_r(x) = w_r \\cdot x`, where :math:`w_r` is the k-th sample weight vector from the underlying distribution
        The probability of choosing an object :math:`x_i` is defined by taking softmax over the
        utility scores of the objects:

        .. math::

            P(x_i \\lvert Q) = \\frac{1}{R} \\sum_{r=1}^R \\frac{exp(U_r(x_i))}{\\sum_{x_j \\in Q} exp(U_r(x_j))}

        The discrete choice for the given query set :math:`Q` is defined as:

        .. math::

            dc(Q) := \\operatorname{argmax}_{x_i \\in Q }  \\; P(x_i \\lvert Q)

        Parameters
        ----------
        n_mixtures: int (range : [2, inf])
            The number of logit models (:math:`R`) which are used to estimate the choice probability
        loss_function : string , {‘categorical_crossentropy’, ‘binary_crossentropy’, ’categorical_hinge’}
            Loss function to be used for the discrete choice decision from the query set
        regularization : string, {‘l1’, ‘l2’}, string
           Regularizer function (L1 or L2) applied to the `kernel` weights matrix
        random_state : int or object
            Numpy random state
        **kwargs
            Keyword arguments for the algorithms

        References
        ----------
            [1] Kenneth E Train. „Discrete choice methods with simulation“. In: Cambridge university press, 2009. Chap Mixed Logit, pp. 153–172.

            [2] Kenneth Train. Qualitative choice analysis. Cambridge, MA: MIT Press, 1986

            [3] Daniel McFadden and Kenneth Train. „Mixed MNL models for discrete response“. In: Journal of applied Econometrics 15.5 (2000), pp. 447–470
        """
        self.loss_function = loss_function
        known_regularization_functions = {"l1", "l2"}
        if regularization not in known_regularization_functions:
            raise ValueError(
                f"Regularization function {regularization} is unknown. Must be one of {known_regularization_functions}"
            )
        self.regularization = regularization
        self.n_mixtures = n_mixtures

    @property
    def model_configuration(self):
        """
            Constructs the dictionary containing the priors for the weight vectors for the model according to the
            regularization function. The parameters are:
                * **weights** : Distribution of the weigh vectors to evaluates the utility of the objects

            For ``l1`` regularization the priors are:

            .. math::

                \\text{mu}_w \\sim \\text{Normal}(\\text{mu}=0, \\text{sd}=5.0) \\\\
                \\text{b}_w \\sim \\text{HalfCauchy}(\\beta=1.0) \\\\
                \\text{weights} \\sim \\text{Laplace}(\\text{mu}=\\text{mu}_w, \\text{b}=\\text{b}_w)

            For ``l2`` regularization the priors are:

            .. math::

                \\text{mu}_w \\sim \\text{Normal}(\\text{mu}=0, \\text{sd}=5.0) \\\\
                \\text{sd}_w \\sim \\text{HalfCauchy}(\\beta=1.0) \\\\
                \\text{weights} \\sim \\text{Normal}(\\text{mu}=\\text{mu}_w, \\text{sd}=\\text{sd}_w)
        """
        if not hasattr(self, "config_"):
            if self.regularization == "l2":
                weight = pm.Normal
                prior = "sd"
            elif self.regularization == "l1":
                weight = pm.Laplace
                prior = "b"
            self.config_ = {
                "weights": [
                    weight,
                    {
                        "mu": (pm.Normal, {"mu": 0, "sd": 5}),
                        prior: (pm.HalfCauchy, {"beta": 1}),
                    },
                ]
            }
            logger.info(
                "Creating model with config {}".format(print_dictionary(self.config_))
            )
        return self.config_

    def construct_model(self, X, Y):
        """
        Constructs the mixed logit model by applying priors on weight vectors **weights** as per
        :meth:`model_configuration`. The probability of choosing the object :math:`x_i` from the query set
        :math:`Q = \\{x_1, \\ldots ,x_n\\}` assuming we draw :math:`R` samples of the weight vectors is:

        .. math::

            P(x_i \\lvert Q) = \\frac{1}{R} \\sum_{r=1}^R \\frac{exp(U_r(x_i))}{\\sum_{x_j \\in Q} exp(U_r(x_j))}

        Parameters
        ----------
        X : numpy array
            (n_instances, n_objects, n_features)
            Feature vectors of the objects
        Y : numpy array
            (n_instances, n_objects)
            Preferences in the form of discrete choices for given objects

        Returns
        -------
         model : pymc3 Model :class:`pm.Model`
        """
        self.trace_ = None
        self.trace_vi = None
        self.loss_function_ = likelihood_dict.get(self.loss_function, None)
        with pm.Model() as self.model:
            self.Xt_ = theano.shared(X)
            self.Yt_ = theano.shared(Y)
            shapes = {"weights": (self.n_object_features_fit_, self.n_mixtures)}
            weights_dict = create_weight_dictionary(self.model_configuration, shapes)
            utility = tt.dot(self.Xt_, weights_dict["weights"])
            self.p_ = tt.mean(ttu.softmax(utility, axis=1), axis=2)
            LogLikelihood(
                "yl", loss_func=self.loss_function_, p=self.p_, observed=self.Yt_
            )
        logger.info("Model construction completed")

    def fit(
        self,
        X,
        Y,
        sampler="variational",
        tune=500,
        draws=500,
        vi_params={
            "n": 20000,
            "method": "advi",
            "callbacks": [CheckParametersConvergence()],
        },
        **kwargs,
    ):
        """
        Fit a mixed logit model on the provided set of queries X and choices Y of those objects. The provided
        queries and corresponding preferences are of a fixed size (numpy arrays). For learning this network
        the categorical cross entropy loss function for each object :math:`x_i \\in Q` is defined as:

        .. math::

            C_{i} =  -y(i)\\log(P_i) \\enspace,

        where :math:`y` is ground-truth discrete choice vector of the objects in the given query set :math:`Q`.
        The value :math:`y(i) = 1` if object :math:`x_i` is chosen else :math:`y(i) = 0`.

        Parameters
        ----------
        X : numpy array (n_instances, n_objects, n_features)
            Feature vectors of the objects
        Y : numpy array (n_instances, n_objects)
            Choices for given objects in the query
        sampler : {‘variational’, ‘metropolis’, ‘nuts’}, string
            The sampler used to estimate the posterior mean and mass matrix from the trace

                * **variational** : Run inference methods to estimate posterior mean and diagonal mass matrix
                * **metropolis** : Use the MAP as starting point and Metropolis-Hastings sampler
                * **nuts** : Use the No-U-Turn sampler
        vi_params : dict
            The parameters for the **variational** inference method
        draws : int
            The number of samples to draw. Defaults to 500. The number of tuned samples are discarded by default
        tune : int
            Number of iterations to tune, defaults to 500. Ignored when using 'SMC'. Samplers adjust
            the step sizes, scalings or similar during tuning. Tuning samples will be drawn in addition
            to the number specified in the `draws` argument, and will be discarded unless
            `discard_tuned_samples` is set to False.
        **kwargs :
            Keyword arguments for the fit function of :meth:`pymc3.fit`or :meth:`pymc3.sample`
        """
        self._pre_fit()
        _n_instances, self.n_objects_fit_, self.n_object_features_fit_ = X.shape
        self.construct_model(X, Y)
        fit_pymc3_model(self, sampler, draws, tune, vi_params, **kwargs)
        return self

    def _predict_scores_fixed(self, X, **kwargs):
        summary = dict(pm.summary(self.trace_)["mean"])
        weights = np.zeros((self.n_object_features_fit_, self.n_mixtures))
        for i, k in product(range(self.n_object_features_fit_), range(self.n_mixtures)):
            weights[i][k] = summary["weights[{},{}]".format(i, k)]
        utility = np.dot(X, weights)
        p = np.mean(npu.softmax(utility, axis=1), axis=2)
        return p



"""An implementation of the scoring module for CmpNet estimators."""

import functools

import torch.nn as nn

from csrank.modules.object_mapping import DenseNeuralNetwork


class RankNetScoring(nn.Module):
    """Map instances to scores with the RankNet approach.

    This approach learns a utility function that considers each object in
    isolation, without taking context into account. It is similar to the
    approach that was introduced in [1]_. This version trains the network based
    on a loss function that evaluates the result (general/discrete choice or
    ranking) as a whole. This differs from the version in [1]_, which proposes
    to evaluate utilities on object pairs and train the network based on that.

    Parameters
    ----------
    n_features: int
        The number of features each object has.

    object_utility_module: pytorch module with one integer parameter
        The module that should be used for utility estimations. Uses a simple
        linear mapping if not specified. You likely want to replace this with
        something more expressive such as a ``DenseNeuralNetwork``. This should
        take the size of the input values as its only parameter. You can use
        ``functools.partial`` if necessary.

    References
    ----------
    .. [1] Burges, C., Shaked, T., Renshaw, E., Lazier, A., Deeds, M.,
    Hamilton, N., & Hullender, G. (2005, August). Learning to rank using
    gradient descent. In Proceedings of the 22nd international conference on
    Machine learning (pp. 89-96).
    """

    def __init__(
        self,
        n_features,
        object_utility_module=functools.partial(
            DenseNeuralNetwork,
            hidden_layers=3,
            units_per_hidden=20,
            ouput_features=1,
        ),
    ):
        super().__init__()
        self.object_utility_module = object_utility_module(n_features)

    def forward(self, instances):
        return self.object_utility_module(instances).squeeze(dim=-1)



"""An implementation of the scoring module for FATE estimators."""

import functools

import torch.nn as nn

from csrank.modules.collection import PairwiseEvaluation
from csrank.modules.instance_reduction import MeanAggregatedUtility
from csrank.modules.object_mapping import CmpNN
from csrank.modules.object_mapping import DenseNeuralNetwork


class CmpNetScoring(nn.Module):
    r"""Map instances to scores with the CmpNet approach.

    This approach was introduced in [1]_. It is very similar to the FETA (first
    evaluate, then aggregate) approach. It uses a ``CmpNet`` for evaluation.
    The network returns pairwise "evidence" that the first object is preferable
    to the second one. That is conceptually the inverse of a pairwise utility
    ("how much does the presence of one object support the utility of
    another?"), but we can ignore that in the implementation since the two are
    equivalently learnable by a neural network. The ``CmpNet`` does not return
    preferences directly but rather "evidence" for preference. That means that
    the preferences are not normalized, which again does not matter for the
    relative scores.

    The evidences of one object against all others are then aggregated by a
    mean reduction, resulting in something similar to a Borda score for each
    object. These scores can be used for ranking and choice tasks. Not that
    this differs from the sorting approach taken in [1]_.

    The main practical difference to the FETA approach is the weight-sharing in
    the evaluation. See the documentation of ``CmpNet`` for details.

    Parameters
    ----------
    n_features: int
        The number of features each object has.
    pairwise_preference_core: instantiated pytorch module
        This module is used for the pairwise evaluations. The original paper
        [1]_ suggests to use a simple linear layer. This module should expect
        an input of shape :math:`(N, 2 \cdot F)` where :math:`F` is the size of
        a feature vector, and produce an output of shape :math:`(N, C)` where
        :math:`C` is the core output size. The output will be further processed
        by the ``CmpNet``.
    core_encoding_size: int
        The size of the embedding that the pairwise preference core should
        produce before the results are aggregated by the final layer. See the
        documentation of ``CmpNet`` for a more detailed understanding.

    References
    ----------
    .. [1] Rigutini, L., Papini, T., Maggini, M., & Scarselli, F. (2011).
    SortNet: Learning to rank by a neural preference function. IEEE
    transactions on neural networks, 22(9), 1368-1380.
    """

    def __init__(
        self,
        n_features,
        pairwise_preference_core=functools.partial(
            DenseNeuralNetwork, hidden_layers=1, units_per_hidden=8
        ),
        core_encoding_size=8,
    ):
        super().__init__()
        self.mean_aggregated_evidence = MeanAggregatedUtility(
            PairwiseEvaluation(
                CmpNN(
                    pairwise_preference_core(2 * n_features, core_encoding_size),
                    core_encoding_size=core_encoding_size,
                )
            ),
            exclude_self_comparison=True,
        )

    def forward(self, instances, **kwargs):
        return self.mean_aggregated_evidence(instances)



"""Loss functions for discrete choice problems."""

import torch


class CategoricalHingeLossMax:
    """Compute the Categorical Hinge Loss.

    This is the "max" aggregated version of CHL, described on page 14/15 of
    [1]_.

    Parameters
    ----------
    scores: 2d tensor
        The predicted scores for each object of each instance.

    true_choice: 2d tensor
        The true choice mask for each instance.

    Returns
    -------
    torch.float
        The total loss, summed over all instances.

    References
    ----------
    .. [1] Pfannschmidt, K., Gupta, P., & Hüllermeier, E. (2019). Learning
    choice functions: Concepts and architectures. arXiv preprint
    arXiv:1901.10860.
    """

    # The argument order is chosen to be compatible with skorch.
    def __call__(self, scores, true_choice):
        """Compute the loss of a scoring in the context of a choice.

        >>> objects = ["a", "b", "c"]
        >>> true_choice_1 = [0, 1, 0]
        >>> scores_1 = [2, 1, 0.5] # non-chosen object "a" has higher score than chosen object "b"
        >>> chl = CategoricalHingeLossMax()
        >>> chl(torch.tensor([scores_1]), torch.tensor([true_choice_1]))
        tensor(2.)

        >>> true_choice_2 = [1, 0, 0]
        >>> scores_2 = [0, 1, 1.5]

        >>> chl = CategoricalHingeLossMax()
        >>> chl(torch.tensor([scores_1, scores_2]), torch.tensor([true_choice_1, true_choice_2]))
        tensor(4.5000)
        """
        # not quite, but dealing with true infintiy is hairy and there should
        # be no practical difference
        infty = 2 ** 32

        # Mask out the chosen scores from the max with a value of -infinity.
        (max_score_not_chosen, _indices) = torch.max(
            scores - true_choice * infty, dim=1
        )
        # Mask out the not-chosen scores from the min with a value of +infinity.
        (min_score_chosen, _indices) = torch.min(
            scores + (1 - true_choice) * infty, dim=1
        )

        hinge = torch.clamp(1 + max_score_not_chosen - min_score_chosen, min=0)
        return hinge.sum()



import functools

import torch.nn as nn

from csrank.modules.object_mapping import DenseNeuralNetwork
from csrank.modules.scoring.feta import FETAScoring
from csrank.objectranking.object_ranker import SkorchObjectRanker
from csrank.rank_losses import HingedRankLoss


class FETAObjectRanker(SkorchObjectRanker):
    """A ranking estimator based on the FETA-Approach.

    See the documentation of :class:`csrank.modules.scoring.FETAScoring` for
    more details.

    Parameters
    ----------
    n_hidden : int
        The number of hidden layers that should be used in the pairwise utility
        module and the zeroth order module (if enabled).

    n_units : int
        The number of units per hidden layer that should be used in the
        pairwise utility module and the zeroth order module (if enabled).

    add_zeroth_order_model : boolean
        Whether or not to add a zeroth order utility model, i.e. a model that
        evaluates an object independent from the context.

    activation : torch activation function (class)
        The activation function that should be used for each layer of the
        pairwise utility module and the zeroth order module (if enabled).

    criterion : torch criterion (class)
        The criterion that is used to evaluate and optimize the module.

    **kwargs : skorch NeuralNet arguments
        All keyword arguments are passed to the constructor of
        ``SkorchObjectRanker``. See the documentation of that class for more
        details.
    """

    def __init__(
        self,
        n_hidden=2,
        n_units=8,
        add_zeroth_order_model=False,
        activation=nn.SELU,
        criterion=HingedRankLoss,
        **kwargs
    ):
        self.n_hidden = n_hidden
        self.n_units = n_units
        self.add_zeroth_order_model = add_zeroth_order_model
        self.activation = activation
        super().__init__(module=FETAScoring, criterion=criterion, **kwargs)

    def _get_extra_module_parameters(self):
        """Return extra parameters that should be passed to the module."""
        params = super()._get_extra_module_parameters()
        params["pairwise_utility_module"] = functools.partial(
            DenseNeuralNetwork,
            hidden_layers=self.n_hidden,
            units_per_hidden=self.n_units,
            activation=self.activation(),
            output_size=1,
        )
        params["zeroth_order_module"] = (
            functools.partial(
                DenseNeuralNetwork,
                hidden_layers=self.n_hidden,
                units_per_hidden=self.n_units,
                activation=self.activation(),
                output_size=1,
            )
            if self.add_zeroth_order_model
            else None
        )
        return params



import logging
import os

from joblib import load
import numpy as np
from sklearn.model_selection import ShuffleSplit
from sklearn.preprocessing import StandardScaler
from sklearn.utils import check_random_state

from csrank.constants import OBJECT_RANKING
from ..dataset_reader import DatasetReader

logger = logging.getLogger(__name__)


class RCVDatasetReader(DatasetReader):
    def __init__(
        self,
        n_instances=10000,
        n_objects=5,
        query_based=False,
        random_state=None,
        **kwargs,
    ):
        super(RCVDatasetReader, self).__init__(
            learning_problem=OBJECT_RANKING, dataset_folder="textual_data", **kwargs
        )
        if n_instances not in [10000]:
            raise ValueError(
                "The number of instances should be in %s", str([100, 1000, 10000])
            )
        if n_objects not in [5, 10]:
            raise ValueError("The number of objects should be in %s", str([5, 10]))

        if not query_based:
            file_name = "rcv1_{}inst_{}obj.pkl".format(n_instances, n_objects)
        else:
            file_name = "rcv1_query-based_{}inst_{}obj.pkl".format(
                n_instances, n_objects
            )

        self.train_file = os.path.join(self.dirname, file_name)
        logger.info(self.train_file)

        self.random_state = check_random_state(random_state)
        self.__load_dataset__()

    def __load_dataset__(self):
        with open(self.train_file, "rb") as file:
            data = load(file)
        orderings = data["orderings"].values
        object_indices = (data["orderings"]["object_index"]).values
        features = (data["features"]).values
        X = []
        for indices in object_indices:
            X.append(features[indices])
        self.X = np.array(X)
        self.Y = np.array(orderings)
        for i, x in enumerate(self.X):
            x = StandardScaler().fit_transform(x)
            self.X[i] = x
        self.__check_dataset_validity__()

    def get_train_test_dataset(self):
        cv_iter = ShuffleSplit(
            n_splits=1, test_size=0.3, random_state=self.random_state
        )
        splits = list(cv_iter.split(self.X))
        return self.splitter(splits)

    def get_dataset_dictionaries(self):
        return self.X, self.Y

    def splitter(self, iter):
        for train_idx, test_idx in iter:
            yield self.X[train_idx], self.Y[train_idx], self.X[test_idx], self.Y[
                test_idx
            ]

    def get_train_test_datasets(self, n_datasets):
        pass

    def get_single_train_test_split(self):
        return self.X, self.Y



import copy
from itertools import product

import numpy as np
from pymc3 import Discrete
from pymc3.distributions.dist_math import bound

from csrank import theano_util as ttu

try:
    import theano
    from theano import tensor as tt
except ImportError:
    from csrank.util import MissingExtraError

    raise MissingExtraError("theano", "probabilistic")


def generate_pairwise_instances(x, choice):
    pairs = np.array(list(product(choice, x)) + list(product(x, choice)))
    n_pairs = len(pairs)
    neg_indices = np.arange(int(n_pairs / 2), n_pairs)
    X1 = pairs[:, 0]
    X2 = pairs[:, 1]
    Y_double = np.ones([n_pairs, 1]) * np.array([1, 0])
    Y_single = np.repeat(1, n_pairs)

    Y_double[neg_indices] = [0, 1]
    Y_single[neg_indices] = 0
    return X1, X2, Y_double, Y_single


def generate_complete_pairwise_dataset(X, Y):
    Y_double = []
    X1 = []
    X2 = []
    Y_single = []
    # Y = np.where(Y==1)
    for x, y in zip(X, Y):
        choice = x[y == 1]
        x = np.delete(x, np.where(y == 1)[0], 0)
        if len(choice) != 0 and len(x) != 0:
            x1, x2, y1, y2 = generate_pairwise_instances(x, choice)
            X1.extend(x1)
            X2.extend(x2)
            Y_double.extend(y1)
            Y_single.extend(y2)
    X1 = np.array(X1)
    X2 = np.array(X2)
    Y_double = np.array(Y_double)
    Y_single = np.array(Y_single)
    X_train = X1 - X2
    return X1, X2, X_train, Y_double, Y_single


def create_weight_dictionary(model_args, shapes):
    weights_dict = dict()
    for key, value in model_args.items():
        prior, params = copy.deepcopy(value)
        for k in params.keys():
            if isinstance(params[k], tuple):
                params[k][1]["name"] = "{}_{}".format(key, k)
                params[k] = params[k][0](**params[k][1])
        params["name"] = key
        params["shape"] = shapes[key]
        weights_dict[key] = prior(**params)
    return weights_dict


def binary_crossentropy(p, y_true):
    axis = 1 if p.ndim > 1 else 0
    return -tt.nnet.binary_crossentropy(p, y_true).mean(axis=axis)


def categorical_crossentropy(p, y_true):
    return -tt.nnet.categorical_crossentropy(p, y_true)


def categorical_hinge(p, y_true):
    pos = tt.sum(y_true * p, axis=-1)
    neg = tt.max((1.0 - y_true) * p, axis=-1)
    return -tt.maximum(0.0, neg - pos + 1.0)


class BinaryCrossEntropyLikelihood(Discrete):
    """
    Categorical log-likelihood.

    The most general discrete distribution.

    .. math:: f(x \\mid p) = p_x

    ========  ===================================
    Support   :math:`x \\in \\{0, 1, \\ldots, |p|-1\\}`
    ========  ===================================

    Parameters
    ----------
    p : array of floats
        p > 0 and the elements of p must sum to 1. They will be automatically
        rescaled otherwise.
    """

    def __init__(self, p, *args, **kwargs):
        super(BinaryCrossEntropyLikelihood, self).__init__(*args, **kwargs)
        self.loss_func = categorical_hinge
        try:
            self.k = tt.shape(p)[-1].tag.test_value
        except AttributeError:
            self.k = tt.shape(p)[-1]
        self.p = tt.as_tensor_variable(p)
        self.mode = tt.argmax(p)

    def random(self, **kwargs):
        return NotImplemented

    def logp(self, value):
        p = self.p
        k = self.k
        a = self.loss_func(p, value)
        p = ttu.normalize(p)
        sum_to1 = theano.gradient.zero_grad(tt.le(abs(tt.sum(p, axis=-1) - 1), 1e-5))

        value_k = tt.argmax(value, axis=1)
        return bound(a, value_k >= 0, value_k <= (k - 1), sum_to1)



import copy
import logging

from csrank.theano_util import normalize

try:
    import pymc3 as pm
    from pymc3 import Discrete
    from pymc3.distributions.dist_math import bound
    from pymc3.variational.callbacks import CheckParametersConvergence
except ImportError:
    from csrank.util import MissingExtraError

    raise MissingExtraError("pymc3", "probabilistic")

try:
    import theano
    from theano import tensor as tt
except ImportError:
    from csrank.util import MissingExtraError

    raise MissingExtraError("theano", "probabilistic")
logger = logging.getLogger(__name__)


def categorical_crossentropy(p, y_true):
    return -tt.nnet.categorical_crossentropy(p, y_true)


def binary_crossentropy(p, y_true):
    axis = 1 if p.ndim > 1 else 0
    return -tt.nnet.binary_crossentropy(p, y_true).mean(axis=axis)


def categorical_hinge(p, y_true):
    pos = tt.sum(y_true * p, axis=-1)
    neg = tt.max((1.0 - y_true) * p, axis=-1)
    return -tt.maximum(0.0, neg - pos + 1.0)


likelihood_dict = {
    "categorical_crossentropy": categorical_crossentropy,
    "binary_crossentropy": binary_crossentropy,
    "categorical_hinge": categorical_hinge,
}


class LogLikelihood(Discrete):
    """
    Categorical log-likelihood.

    The most general discrete distribution.

    .. math:: f(x \\mid p) = p_x

    ========  ===================================
    Support   :math:`x \\in \\{0, 1, \\ldots, |p|-1\\}`
    ========  ===================================

    Parameters
    ----------
    p : array of floats
        p > 0 and the elements of p must sum to 1. They will be automatically
        rescaled otherwise.
    """

    def __init__(self, loss_func, p, *args, **kwargs):
        super(LogLikelihood, self).__init__(*args, **kwargs)
        if loss_func is None:
            loss_func = categorical_crossentropy
        self.loss_func = loss_func
        try:
            self.k = tt.shape(p)[-1].tag.test_value
        except AttributeError:
            self.k = tt.shape(p)[-1]
        self.p = tt.as_tensor_variable(p)
        self.mode = tt.argmax(p)

    def random(self, **kwargs):
        return NotImplemented

    def logp(self, value):
        p = self.p
        k = self.k
        a = self.loss_func(p, value)
        p = normalize(p)
        sum_to1 = theano.gradient.zero_grad(tt.le(abs(tt.sum(p, axis=-1) - 1), 1e-5))

        value_k = tt.argmax(value, axis=1)
        return bound(a, value_k >= 0, value_k <= (k - 1), sum_to1)


def create_weight_dictionary(model_args, shapes):
    weights_dict = dict()
    for key, value in model_args.items():
        prior, params = copy.deepcopy(value)
        for k in params.keys():
            if isinstance(params[k], tuple):
                params[k][1]["name"] = "{}_{}".format(key, k)
                params[k] = params[k][0](**params[k][1])
        params["name"] = key
        params["shape"] = shapes[key]
        weights_dict[key] = prior(**params)
    return weights_dict


def fit_pymc3_model(self, sampler, draws, tune, vi_params, **kwargs):
    callbacks = vi_params.get("callbacks", [])
    for i, c in enumerate(callbacks):
        if isinstance(c, CheckParametersConvergence):
            params = c.__dict__
            params.pop("_diff")
            params.pop("prev")
            params.pop("ord")
            params["diff"] = "absolute"
            callbacks[i] = CheckParametersConvergence(**params)
    if sampler == "variational":
        with self.model:
            try:
                self.trace_ = pm.sample(chains=2, cores=8, tune=5, draws=5)
                vi_params["start"] = self.trace_[-1]
                self.trace_vi_ = pm.fit(**vi_params)
                self.trace_ = self.trace_vi_.sample(draws=draws)
            except Exception as e:
                if hasattr(e, "message"):
                    message = e.message
                else:
                    message = e
                logger.error(message)
                self.trace_vi_ = None
        if self.trace_vi_ is None and self.trace_ is None:
            with self.model:
                logger.info(
                    "Error in vi ADVI sampler using Metropolis sampler with draws {}".format(
                        draws
                    )
                )
                self.trace = pm.sample(
                    chains=1, cores=4, tune=20, draws=20, step=pm.NUTS()
                )
    elif sampler == "metropolis":
        with self.model:
            start = pm.find_MAP()
            self.trace_ = pm.sample(
                chains=2,
                cores=8,
                tune=tune,
                draws=draws,
                **kwargs,
                step=pm.Metropolis(),
                start=start,
            )
    else:
        with self.model:
            self.trace_ = pm.sample(
                chains=2, cores=8, tune=tune, draws=draws, **kwargs, step=pm.NUTS()
            )



# https://github.com/sqlalchemyorg/zimports/issues/12
# fmt: off
from .depth_dataset_reader import DepthDatasetReader
from .image_dataset_reader import ImageDatasetReader
from .letor_listwise_object_ranking_dataset_reader import LetorListwiseObjectRankingDatasetReader
from .neural_sentence_ordering_reader import SentenceOrderingDatasetReader
from .object_ranking_data_generator import ObjectRankingDatasetGenerator
from .rcv_dataset_reader import RCVDatasetReader
from .sushi_object_ranking_dataset_reader import SushiObjectRankingDatasetReader
from .tag_genome_object_ranking_dataset_reader import TagGenomeObjectRankingDatasetReader
# fmt: on

__all__ = [
    "DepthDatasetReader",
    "ImageDatasetReader",
    "LetorListwiseObjectRankingDatasetReader",
    "SentenceOrderingDatasetReader",
    "ObjectRankingDatasetGenerator",
    "RCVDatasetReader",
    "SushiObjectRankingDatasetReader",
    "TagGenomeObjectRankingDatasetReader",
]



from .baseline import AllPositive
from .cmpnet_choice import CmpNetChoiceFunction
from .fate_choice import FATEChoiceFunction
from .feta_choice import FETAChoiceFunction
from .generalized_linear_model import GeneralizedLinearModel
from .pairwise_choice import PairwiseSVMChoiceFunction
from .ranknet_choice import RankNetChoiceFunction

__all__ = [
    "AllPositive",
    "CmpNetChoiceFunction",
    "FATEChoiceFunction",
    "FETAChoiceFunction",
    "GeneralizedLinearModel",
    "PairwiseSVMChoiceFunction",
    "RankNetChoiceFunction",
]



"""Loss functions for ranking choice problems."""

import torch


class HingedRankLoss:
    """Compute the pairwise loss between two lists of rankings.

    Assumes the true ranking is represented by a permutation array (list of
    indices). The scores are floats, which should be high for low ranking
    indices.

    Their relative order of the scores (i.e. their reversed argsort) and their
    difference is important for the resulting loss. Even scores that result in
    a correct ranking will be penalized if the scores of two elements differ by
    less than 1. For example if we have objects

    >>> objects = ["a", "b", "c"]

    then the true ranking (permutation array)

    >>> ranking = [0, 2, 1]

    and the scores

    >>> scores = [0, -10, -5]

    would equivalently specify the ranking ["a, "c", "b"] and consequently

    >>> hinged_rank_loss = HingedRankLoss()
    >>> hinged_rank_loss(torch.tensor([scores]), torch.tensor([ranking]))
    tensor(0.)

    A rescaled scoring with an insufficient gap would lead to a non-zero loss:

    >>> scores = [0, -1, -0.5]
    >>> hinged_rank_loss(torch.tensor([scores]), torch.tensor([ranking]))
    tensor(0.3333)

    The ranking

    >>> ranking = [0, 2, 1]

    is not matched by the scoring

    >>> scores = [-2, 0, -1]

    and thus

    >>> hinged_rank_loss(torch.tensor([scores]), torch.tensor([ranking]))
    tensor(2.3333)

    Parameters
    ----------
    comparison_scores: 2d tensor
        The predicted scores for each object of each instance.

    true_rankings: 2d tensor
        The true rankings, represented as a permutation. The first element of a
        permutation contains the index to which the first element should be
        moved.

    Returns
    -------
    torch.float
        The total loss, summed over all instances.
    """

    # The argument order is chosen to be compatible with skorch.
    def __call__(self, comparison_scores, true_rankings):
        # 2d matrix which is 1 if the row-element *should* be ranked higher than the column element
        mask = true_rankings[:, :, None] > true_rankings[:, None]
        # How much higher/lower the elements are actually ranked. First create
        # new dimensions (at the element/instance level), then rely on
        # broadcasting to compute the difference.
        # Negated because higher scores imply lower rankings.
        diff = -(comparison_scores[:, :, None] - comparison_scores[:, None])
        self.diff = diff
        hinge = torch.clamp(mask * (1 - diff), min=0)
        n = torch.sum(mask, axis=(1, 2))
        losses = torch.true_divide(torch.sum(hinge, axis=(1, 2)), n)
        return losses.sum()



from importlib.metadata import version

from .choicefunction import *  # noqa
from .core import *  # noqa
from .discretechoice import *  # noqa
from .objectranking import *  # noqa

# We should re-evaluate if we really want to re-export everything here and then
# use __all__ properly.

__version__ = version(__name__)



import numpy as np
from pymc3.variational.callbacks import CheckParametersConvergence
import pytest
import torch
from torch import optim

from csrank.choicefunction import CmpNetChoiceFunction
from csrank.choicefunction import FATEChoiceFunction
from csrank.choicefunction import FETAChoiceFunction
from csrank.choicefunction import GeneralizedLinearModel
from csrank.choicefunction import PairwiseSVMChoiceFunction
from csrank.choicefunction import RankNetChoiceFunction
from csrank.constants import CMPNET_CHOICE
from csrank.constants import FATE_CHOICE
from csrank.constants import FETA_CHOICE
from csrank.constants import GLM_CHOICE
from csrank.constants import RANKNET_CHOICE
from csrank.constants import RANKSVM_CHOICE
from csrank.metrics_np import auc_score
from csrank.metrics_np import f1_measure
from csrank.metrics_np import instance_informedness
from csrank.metrics_np import subset_01_loss
from csrank.util import metrics_on_predictions

choice_metrics = {
    "F1Score": f1_measure,
    "Informedness": instance_informedness,
    "AucScore": auc_score,
}


def get_vals(values):
    return dict(zip(choice_metrics.keys(), values))


skorch_common_args = {
    "max_epochs": 100,
    "optimizer": optim.SGD,
    "optimizer__lr": 1e-3,
    "optimizer__momentum": 0.9,
    "optimizer__nesterov": True,
    # We evaluate the estimators in-sample. These tests are just small
    # sanity checks, so overfitting is okay here.
    "train_split": None,
}

choice_functions = {
    CMPNET_CHOICE: (
        CmpNetChoiceFunction,
        {"n_hidden": 2, "n_units": 8, **skorch_common_args},
        get_vals([0.8728, 0.8864, 1.0]),
    ),
    GLM_CHOICE: (GeneralizedLinearModel, {}, get_vals([0.9567, 0.9955, 1.0])),
    RANKSVM_CHOICE: (PairwiseSVMChoiceFunction, {}, get_vals([0.9522, 0.9955, 1.0])),
    FATE_CHOICE: (
        FATEChoiceFunction,
        {
            "n_hidden_joint_layers": 1,
            "n_hidden_set_layers": 1,
            "n_hidden_joint_units": 5,
            "n_hidden_set_units": 5,
            **skorch_common_args,
        },
        get_vals([0.7177, 0.3119, 1.0]),
    ),
    FETA_CHOICE: (
        FETAChoiceFunction,
        {
            "n_hidden": 1,
            "n_units": 8,
            "add_zeroth_order_model": False,
            **skorch_common_args,
        },
        get_vals([0.8759, 0.8855, 1.0]),
    ),
    FETA_CHOICE
    + "_zeroth_order_model": (
        FETAChoiceFunction,
        {
            "n_hidden": 1,
            "n_units": 8,
            "add_zeroth_order_model": True,
            **skorch_common_args,
        },
        get_vals([0.9368, 0.9617, 1.0]),
    ),
    RANKNET_CHOICE: (
        RankNetChoiceFunction,
        {"n_hidden": 2, "n_units": 8, **skorch_common_args},
        get_vals([0.9202, 0.9198, 1.0]),
    ),
}


@pytest.fixture(scope="module")
def trivial_choice_problem():
    random_state = np.random.RandomState(42)
    # pytorch uses 32 bit floats by default. That should be precise enough and
    # makes it easier to use pytorch and non-pytorch estimators interchangeably.
    x = random_state.randn(200, 5, 1).astype(np.float32)
    # The pytorch estimators expect booleans to be encoded as a 32 bit float
    # (1.0 for True, 0.0 for false).
    y_true = np.array(x.squeeze(axis=-1) > np.mean(x), dtype=np.float32)
    return x, y_true


@pytest.mark.parametrize("name", list(choice_functions.keys()))
def test_choice_function_fixed(trivial_choice_problem, name):
    np.random.seed(123)
    # Pytorch does not guarantee full reproducibility in different settings
    # [1]. This may become a problem in the test suite, in which case we should
    # increase the tolerance. These are only "sanity checks" on small data sets
    # anyway and the exact values do not mean much here.
    # [1] https://pytorch.org/docs/stable/notes/randomness.html
    torch.manual_seed(123)
    # Trade off performance for better reproducibility.
    torch.use_deterministic_algorithms(True)
    x, y = trivial_choice_problem
    choice_function = choice_functions[name][0]
    params, accuracies = choice_functions[name][1], choice_functions[name][2]
    learner = choice_function(**params)
    if name == GLM_CHOICE:
        learner.fit(
            x,
            y,
            vi_params={
                "n": 100,
                "method": "advi",
                "callbacks": [CheckParametersConvergence()],
            },
        )
    else:
        learner.fit(x, y)

    s_pred = learner.predict_scores(x)
    y_pred = learner.predict_for_scores(s_pred)
    y_pred_2 = learner.predict(x)
    rtol = 1e-2
    atol = 5e-2
    assert np.isclose(
        0.0, subset_01_loss(y_pred, y_pred_2), rtol=rtol, atol=atol, equal_nan=False
    )
    for key, value in accuracies.items():
        metric = choice_metrics[key]
        if metric in metrics_on_predictions:
            pred_loss = metric(y, y_pred)
        else:
            pred_loss = metric(y, s_pred)
        assert np.isclose(value, pred_loss, rtol=rtol, atol=atol, equal_nan=False)



import itertools as iter
import sys

import numpy as np
from sklearn.metrics import f1_score
from sklearn.preprocessing import StandardScaler

try:
    import pandas as pd
except ImportError:
    from csrank.util import MissingExtraError

    raise MissingExtraError("pandas", "data")


def strongly_connected_components(graph):
    """Find the strongly connected components in a graph using
    Tarjan's algorithm.
    # Taken from http://www.logarithmic.net/pfh-files/blog/01208083168/sort.py

    graph should be a dictionary mapping node names to
    lists of successor nodes.
    """

    result = []
    stack = []
    low = {}

    def visit(node):
        if node in low:
            return

        num = len(low)
        low[node] = num
        stack_pos = len(stack)
        stack.append(node)

        for successor in graph[node]:
            visit(successor)
            low[node] = min(low[node], low[successor])

        if num == low[node]:
            component = tuple(stack[stack_pos:])
            del stack[stack_pos:]
            result.append(component)
            for item in component:
                low[item] = len(graph)

    for node in graph:
        visit(node)

    return result


def create_graph_pairwise_matrix(matrix):
    n_objects = matrix.shape[0]
    graph = {key: [] for key in np.arange(n_objects)}
    for i, j in iter.combinations(np.arange(n_objects), 2):
        p_ij = matrix[i, j]
        p_ji = matrix[j, i]
        if p_ij > p_ji:
            graph[j].append(i)
        if p_ij < p_ji:
            graph[i].append(j)
        if p_ij == p_ji:
            graph[j].append(i)
            graph[i].append(j)
    return graph


def create_pairwise_prob_matrix(n_objects):
    # Create a non-transitive pairwise probability matrix for n_objects*n_objects
    non_transitive = False
    while not non_transitive:
        pairwise_prob = np.zeros([n_objects, n_objects])
        for i, j in iter.combinations(np.arange(n_objects), 2):
            pairwise_prob[i, j] = np.random.rand(1)[0]
            pairwise_prob[j, i] = 1.0 - pairwise_prob[i, j]
        for comp in strongly_connected_components(
            create_graph_pairwise_matrix(pairwise_prob)
        ):
            if len(comp) >= 3:
                non_transitive = True
                break
    return pairwise_prob


def quicksort(arr, matrix):
    # Apply the quick sort algorithm for the given set of objects and produces an ordering based on provided pairwise matrix
    if len(arr) < 2:
        return arr
    else:
        pivot = np.random.choice(arr, 1)[0]
        arr.remove(pivot)
        right = [i for i in arr if matrix[pivot, i] == 1]
        left = [i for i in arr if matrix[pivot, i] == 0]
        return quicksort(left, matrix) + [pivot] + quicksort(right, matrix)


def get_similarity_matrix(file_path):
    data_frame = pd.read_csv(file_path)
    similarity_dictionary = data_frame.set_index("col_major_index")[
        "similarity"
    ].to_dict()
    return similarity_dictionary


def get_key_for_indices(idx1, idx2):
    return str(tuple(sorted([idx1, idx2])))


def distance_metric_multilabel(x_labels, y_labels, X, Y):
    similarity = f1_score(x_labels, y_labels, average="macro")
    similarity = np.dot(X, Y) / (np.linalg.norm(X) * np.linalg.norm(Y)) + similarity
    return similarity


def print_no_newline(i, total):
    sys.stdout.write("Iterations: {}  out of {} \r".format(i, total))
    sys.stdout.flush()


def standardize_features(x_train, x_test):
    standardize = Standardize()
    x_train = standardize.fit_transform(x_train)
    x_test = standardize.transform(x_test)
    return x_train, x_test


class Standardize(object):
    def __init__(self, scalar=StandardScaler):
        self.scalar = scalar()
        self.n_features = None

    def fit(self, X):
        if isinstance(X, dict):
            n_features = X[list(X.keys())[0]].shape[-1]
            X = []
            for x in X.values():
                X.extend(x.reshape(-1, n_features))
            X = np.array(X)
            self.scalar.fit(X)
        if isinstance(X, (np.ndarray, np.generic)):
            n_features = X.shape[-1]
            X = X.reshape(-1, n_features)
            self.scalar.fit(X)
        self.n_features = n_features

    def transform(self, X):
        if isinstance(X, dict):
            for n in X.keys():
                X[n] = X[n].reshape(-1, self.n_features)
                X[n] = self.scalar.transform(X[n])
                X[n] = X[n].reshape(-1, n, self.n_features)
        if isinstance(X, (np.ndarray, np.generic)):
            n_objects = X.shape[-2]
            X = X.reshape(-1, self.n_features)
            X = self.scalar.transform(X)
            X = X.reshape(-1, n_objects, self.n_features)
        return X

    def fit_transform(self, X):
        self.fit(X)
        X = self.transform(X)
        return X



# https://github.com/sqlalchemyorg/zimports/issues/12
# fmt: off
from .discrete_choice_data_generator import DiscreteChoiceDatasetGenerator
from .expedia_discrete_choice_dataset_reader import ExpediaDiscreteChoiceDatasetReader
from .letor_listwise_discrete_choice_dataset_reader import LetorListwiseDiscreteChoiceDatasetReader
from .letor_ranking_discrete_choice_dataset_reader import LetorRankingDiscreteChoiceDatasetReader
from .mnist_discrete_choice_dataset_reader import MNISTDiscreteChoiceDatasetReader
from .sushi_discrete_choice_dataset_reader import SushiDiscreteChoiceDatasetReader
from .tag_genome_discrete_choice_dataset_reader import TagGenomeDiscreteChoiceDatasetReader
# fmt: on

__all__ = [
    "DiscreteChoiceDatasetGenerator",
    "ExpediaDiscreteChoiceDatasetReader",
    "LetorListwiseDiscreteChoiceDatasetReader",
    "LetorRankingDiscreteChoiceDatasetReader",
    "MNISTDiscreteChoiceDatasetReader",
    "SushiDiscreteChoiceDatasetReader",
    "TagGenomeDiscreteChoiceDatasetReader",
]



import functools

import torch.nn as nn

from csrank.modules.object_mapping import DenseNeuralNetwork
from csrank.modules.scoring.cmpnet import CmpNetScoring
from csrank.objectranking.object_ranker import SkorchObjectRanker
from csrank.rank_losses import HingedRankLoss


class CmpNetObjectRanker(SkorchObjectRanker):
    """A object ranking estimator based on the CmpNet-Approach.

    See the documentation of :class:`csrank.modules.scoring.CmpNetScoring` for
    more details.

    Parameters
    ----------
    n_hidden : int
        The number of hidden layers that should be used in the pairwise utility
        module and the zeroth order module (if enabled).

    n_units : int
        The number of units per hidden layer that should be used in the
        pairwise utility module and the zeroth order module (if enabled).

    activation : torch activation function (class)
        The activation function that should be used for each layer of the
        comparative network.

    criterion : torch criterion (class)
        The criterion that is used to evaluate and optimize the module.

    **kwargs : skorch NeuralNet arguments
        All keyword arguments are passed to the constructor of
        ``SkorchChoiceFunction``. See the documentation of that class for more
        details.
    """

    def __init__(
        self,
        n_hidden=2,
        n_units=8,
        criterion=HingedRankLoss,
        activation=nn.SELU,
        **kwargs
    ):
        self.n_hidden = n_hidden
        self.n_units = n_units
        self.activation = activation
        super().__init__(module=CmpNetScoring, criterion=criterion, **kwargs)

    def _get_extra_module_parameters(self):
        """Return extra parameters that should be passed to the module."""
        params = super()._get_extra_module_parameters()
        params["pairwise_preference_core"] = functools.partial(
            DenseNeuralNetwork,
            hidden_layers=self.n_hidden,
            units_per_hidden=self.n_units,
            activation=self.activation(),
        )
        params["core_encoding_size"] = self.n_hidden
        return params



from abc import ABCMeta
from abc import abstractmethod
from itertools import combinations
import logging
import os

import numpy as np
from sklearn.model_selection import train_test_split
from sklearn.utils import check_random_state

from csrank.dataset_reader.dataset_reader import DatasetReader
from .util import get_key_for_indices
from .util import get_similarity_matrix
from .util import standardize_features

try:
    import pandas as pd
except ImportError:
    from csrank.util import MissingExtraError

    raise MissingExtraError("pandas", "data")

MOVIE_ID = "movieId"
TAG_ID = "tagId"
TAG = "tag"
TAG_POPULARITY = "tagpopularity"
DOC_FREQUENCY = "docfrequency"
RELEVANCE = "relevance"

logger = logging.getLogger(__name__)


class TagGenomeDatasetReader(DatasetReader, metaclass=ABCMeta):
    def __init__(
        self,
        n_train_instances=10000,
        n_test_instances=10000,
        n_objects=5,
        random_state=None,
        standardize=True,
        **kwargs,
    ):
        super(TagGenomeDatasetReader, self).__init__(
            dataset_folder="movie_lens", **kwargs
        )

        self.movies_file = os.path.join(self.dirname, "movies.csv")
        genome_scores = pd.read_csv(os.path.join(self.dirname, "genome-scores.csv"))
        tags_applies = pd.read_csv(os.path.join(self.dirname, "tags.csv"))

        self.tags_info_file = os.path.join(self.dirname, "genome-tags.csv")
        genome_tags = pd.read_csv(self.tags_info_file)
        movies_df = pd.read_csv(self.movies_file)
        self.similarity_matrix_file = os.path.join(
            self.dirname, "similarity_matrix.csv"
        )
        self.standardize = standardize
        self.n_objects = n_objects
        self.n_features = np.array(genome_tags[TAG_ID]).shape[0]
        self.n_test_instances = n_test_instances
        self.n_train_instances = n_train_instances
        self.random_state = check_random_state(random_state)
        if not os.path.isfile(self.similarity_matrix_file):
            self.__load_dataset__(genome_scores, genome_tags, tags_applies, movies_df)

        logger.info("Loading similarity matrix")
        self.similarity_matrix = get_similarity_matrix(self.similarity_matrix_file)
        logger.info("Done loading similarity matrix")
        self.weights = np.log(np.array(genome_tags[TAG_POPULARITY])) / np.log(
            np.array(genome_tags[DOC_FREQUENCY])
        )
        self.movies_df = pd.read_csv(self.movies_file)
        self.n_movies = len(self.movies_df)
        self.movie_features = self.movies_df.as_matrix()[:, 3:].astype(float)
        self.tags_info_df = pd.read_csv(self.tags_info_file)
        logger.info("Done creating the complete dataset")

    def __load_dataset__(self, genome_scores, genome_tags, tags_applies, movies_df):
        tags_applies[TAG] = tags_applies[TAG].str.lower()
        values_with_pop = []
        for tid, t in genome_tags.as_matrix():
            tags_a = tags_applies.loc[tags_applies[TAG] == t.lower()]
            tags_total = np.array(tags_a[MOVIE_ID])
            values_with_pop.append([tid, t, len(tags_total)])
            logger.info("Tag popularity for tag {} is: {}".format(t, len(tags_total)))
            if len(tags_total) == 0:
                logger.info((tid, t))
                logger.info(
                    "Tag popularity for tag {} is: {}".format(t, len(tags_total))
                )
                values_with_pop.append([tid, t, 2])
        cols = list(genome_tags.columns)
        cols.append(TAG_POPULARITY)
        genome_tags = pd.DataFrame(values_with_pop, columns=cols)
        doc_freq = []
        for i, tag in enumerate(genome_tags.values):
            df = genome_scores.loc[genome_scores["tagId"] == tag[0]]
            freq = np.sum(np.array(df["relevance"]) > 0.5)
            if freq == 0 or freq == 1:
                freq = 2
                logger.info("Document frequency for tag {} is: zero".format(tag))
            doc_freq.append(freq)
        genome_tags[DOC_FREQUENCY] = doc_freq
        genome_tags.to_csv(self.tags_info_file, index=False)
        logger.info("Done loading the tag popularity and doc frequency for the tags")
        self.weights = np.log(np.array(genome_tags[TAG_POPULARITY])) / np.log(
            np.array(genome_tags[DOC_FREQUENCY])
        )

        objects = []
        tags_rel = genome_scores.as_matrix()
        movie_ids = np.unique(np.array(genome_scores[MOVIE_ID]))
        for i, movie_id in enumerate(movie_ids):
            a = i * self.n_features
            b = (i + 1) * self.n_features
            objects.append(tags_rel[a:b, 2])
        objects = np.array(objects)
        movies_df = movies_df[movies_df[MOVIE_ID].isin(movie_ids)]

        for i, tag in enumerate(genome_tags.values):
            movies_df[tag[1]] = objects[:, i]
        movies_df.to_csv(self.movies_file, index=False)

        logger.info("Done loading the features for the movies")

        num_of_movies = movie_ids.shape[0]
        combinations_list = np.array(list(combinations(range(num_of_movies), 2)))
        similarity_matrix = dict()

        features = movies_df.as_matrix()[:, 3:]

        for i, j in combinations_list:
            similarity_matrix[get_key_for_indices(i, j)] = weighted_cosine_similarity(
                self.weights
            )(features[i], features[j])
            logger.info(
                "Calculating similarity {},{}, {}".format(
                    i, j, similarity_matrix[get_key_for_indices(i, j)]
                )
            )

        for i in range(num_of_movies):
            similarity_matrix[get_key_for_indices(i, i)] = 1.0

        series = pd.Series(similarity_matrix)
        matrix_df = pd.DataFrame(
            {"col_major_index": series.index, "similarity": series.values}
        )
        matrix_df.to_csv(self.similarity_matrix_file, index=False)
        logger.info(
            "Done calculating the similarity matrix stored at: {}".format(
                self.similarity_matrix_file
            )
        )

    @abstractmethod
    def make_nearest_neighbour_dataset(self, n_instances, n_objects, seed, **kwargs):
        logger.info(
            "For instances {} objects {}, seed {}".format(n_instances, n_objects, seed)
        )
        random_state = check_random_state(seed)
        X = np.empty((n_instances, n_objects, self.n_features), dtype=float)
        scores = np.empty((n_instances, n_objects), dtype=float)
        for i in range(n_instances):
            subset = random_state.choice(self.n_movies, size=n_objects, replace=False)
            X[i] = self.movie_features[subset]
            D = np.ones((n_objects, n_objects), dtype=float)
            for j, k in combinations(np.arange(n_objects), 2):
                D[j, k] = D[k, j] = self.similarity_matrix[
                    get_key_for_indices(subset[j], subset[k])
                ]
            sum_dist = D.mean(axis=0)
            medoid = np.argmax(sum_dist)
            scores[i] = D[medoid]
        return X, scores

    def get_genre_tag_id(self):
        genres = []
        for g in self.movies_df["genres"]:
            arr = g.lower().split("|")
            genres.extend(arr)
        genres = list(set(genres))
        genre_tag_id = []
        for row in self.tags_info_df.sort_values(
            "tagpopularity", ascending=False
        ).values:
            if row[1].lower() in genres or len(genre_tag_id) < 18:
                genre_tag_id.append(row[0] - 1)
        return np.array(genre_tag_id)

    @abstractmethod
    def make_critique_fit_dataset(
        self, n_instances, n_objects, seed, direction, **kwargs
    ):
        logger.info(
            "For instances {} objects {}, seed {}, direction {}".format(
                n_instances, n_objects, seed, direction
            )
        )
        random_state = check_random_state(seed)
        X = []
        scores = []
        length = int(n_instances / self.n_movies) + 1
        popular_tags = self.get_genre_tag_id()
        for i, feature in enumerate(self.movie_features):
            if direction == 1:
                quartile_tags = np.where(
                    np.logical_and(feature >= 1 / 3, feature < 2 / 3)
                )[0]
            else:
                quartile_tags = np.where(feature > 1 / 2)[0]
            if len(quartile_tags) < length:
                quartile_tags = popular_tags
            tag_ids = random_state.choice(quartile_tags, size=length)
            distances = [
                self.similarity_matrix[get_key_for_indices(i, j)]
                for j in range(self.n_movies)
            ]
            distances = np.array(distances)
            critique_d = critique_dist(
                feature, self.movie_features, tag_ids, direction=direction, relu=False
            )
            critique_fit = np.multiply(critique_d, distances)
            orderings = np.argsort(critique_fit, axis=-1)[:, ::-1][
                :, 0 : (n_objects - 1)
            ]
            orderings = np.append(
                orderings, np.zeros(length, dtype=int)[:, None] + i, axis=1
            )
            to_remove = []
            for j, ordering in enumerate(orderings):
                random_state.shuffle(ordering)
                if len(np.unique(critique_fit[j, ordering])) != n_objects:
                    to_remove.append(j)
            cf_scores = critique_fit[np.arange(length)[:, None], orderings]
            movies = self.movie_features[orderings]
            if len(to_remove) != 0:
                cf_scores = np.delete(cf_scores, to_remove, 0)
                movies = np.delete(movies, to_remove, 0)
            scores.extend(cf_scores)
            X.extend(movies)
        X = np.array(X)
        scores = np.array(scores)
        indices = random_state.choice(X.shape[0], n_instances, replace=False)
        X = X[indices, :, :]
        scores = scores[indices, :]
        return X, scores

    def get_dataset_dictionaries(self, lengths=[5, 6]):
        x_train = dict()
        y_train = dict()
        x_test = dict()
        y_test = dict()
        for n_obj in lengths:
            seed = self.random_state.randint(2 ** 32, dtype="uint32")
            total_instances = self.n_test_instances + self.n_train_instances
            X, Y = self.dataset_function(total_instances, n_obj, seed=seed)
            x_1, x_2, y_1, y_2 = train_test_split(
                X, Y, random_state=self.random_state, test_size=self.n_test_instances
            )
            if self.standardize:
                x_1, x_2 = standardize_features(x_1, x_2)
            x_train[n_obj], x_test[n_obj], y_train[n_obj], y_test[n_obj] = (
                x_1,
                x_2,
                y_1,
                y_2,
            )
        logger.info("Done")
        return x_train, y_train, x_test, y_test

    def get_train_test_datasets(self, n_datasets=5):
        splits = np.array(n_datasets)
        logger.info("Done")
        return self.splitter(splits)

    def get_single_train_test_split(self):
        seed = self.random_state.randint(2 ** 32, dtype="uint32")
        total_instances = self.n_test_instances + self.n_train_instances
        self.X, self.Y = self.dataset_function(
            total_instances, self.n_objects, seed=seed
        )
        self.__check_dataset_validity__()
        x_train, x_test, y_train, y_test = train_test_split(
            self.X,
            self.Y,
            random_state=self.random_state,
            test_size=self.n_test_instances,
        )
        if self.standardize:
            x_train, x_test = standardize_features(x_train, x_test)
        logger.info("Done")
        return x_train, y_train, x_test, y_test

    def splitter(self, iter):
        for i in iter:
            seed = self.random_state.randint(2 ** 32, dtype="uint32") + i
            total_instances = self.n_test_instances + self.n_train_instances
            X, Y = self.dataset_function(total_instances, self.n_objects, seed=seed)
            x_train, x_test, y_train, y_test = train_test_split(
                X, Y, random_state=self.random_state, test_size=self.n_test_instances
            )
            if self.standardize:
                x_train, x_test = standardize_features(x_train, x_test)

            yield x_train, y_train, x_test, y_test


def weighted_cosine_similarity(weights):
    def distance_function(x, y):
        denominator = np.sqrt(np.sum(weights * x * x)) * np.sqrt(
            np.sum(weights * y * y)
        )
        similarity = np.sum(weights * x * y) / denominator
        return similarity

    return distance_function


def critique_dist(ic, irs, tag_ids, direction=-1, relu=True):
    distances = ((irs[:, tag_ids] - ic[tag_ids]) * direction).T
    if relu:
        distances = np.maximum(np.zeros_like(distances), distances)
    return distances



from functools import partial

import numpy as np
from scipy.stats import spearmanr
from sklearn.metrics import average_precision_score
from sklearn.metrics import f1_score
from sklearn.metrics import hamming_loss
from sklearn.metrics import precision_score
from sklearn.metrics import recall_score
from sklearn.metrics import roc_auc_score
from sklearn.metrics import zero_one_loss

from csrank.numpy_util import scores_to_rankings

__all__ = [
    "spearman_correlation_for_scores_np",
    "kendalls_tau_for_scores_np",
    "zero_one_accuracy_for_scores_np",
    "zero_one_rank_loss_for_scores_ties_np",
    "zero_one_accuracy_np",
    "zero_one_rank_loss_for_scores_np",
    "auc_score",
    "instance_informedness",
    "f1_measure",
    "recall",
    "hamming",
    "average_precision",
    "precision",
    "subset_01_loss",
    "spearman_correlation_for_scores_scipy",
    "topk_categorical_accuracy_np",
    "categorical_accuracy_np",
    "make_ndcg_at_k_loss_np",
]


def spearman_correlation_for_scores_np(y_true, s_pred):
    y_pred = scores_to_rankings(s_pred)
    rho = []
    n_objects = y_true.shape[1]
    denominator = n_objects * (n_objects ** 2 - 1)

    for r1, r2 in zip(y_true, y_pred):
        if len(np.unique(r2)) == len(r2):
            s = 1 - (6 * np.sum((r1 - r2) ** 2) / denominator)
            rho.append(s)
        else:
            rho.append(np.nan)
    return np.nanmean(np.array(rho))


def spearman_correlation_for_scores_scipy(y_true, s_pred):
    y_pred = scores_to_rankings(s_pred)
    rho = []
    for r1, r2 in zip(y_true, y_pred):
        s = spearmanr(r1, r2)[0]
        rho.append(s)
    return np.nanmean(np.array(rho))


def kendalls_tau_for_scores_np(y_true, s_pred):
    return 1.0 - 2.0 * zero_one_rank_loss_for_scores_ties_np(y_true, s_pred)


def zero_one_accuracy_for_scores_np(y_true, s_pred):
    y_pred = scores_to_rankings(s_pred)
    acc = np.sum(np.all(np.equal(y_true, y_pred), axis=1)) / y_pred.shape[0]
    return acc


def zero_one_accuracy_np(y_true, y_pred):
    acc = np.sum(np.all(np.equal(y_true, y_pred), axis=1)) / y_pred.shape[0]
    return acc


def make_ndcg_at_k_loss_np(k=5):
    def ndcg(y_true, y_pred):
        n_instances, n_objects = y_true.shape
        relevance = np.power(2.0, ((n_objects - y_true) * 60) / n_objects) - 1.0
        relevance_pred = np.power(2.0, ((n_objects - y_pred) * 60) / n_objects) - 1.0

        log_term = np.log(np.arange(k, dtype="float32") + 2.0) / np.log(2.0)

        # Calculate ideal dcg:
        top_t = np.argsort(relevance, axis=1)[:, ::-1][:, :k]
        toprel = relevance[np.arange(n_instances)[:, None], top_t]
        idcg = np.sum(toprel / log_term, axis=-1, keepdims=True)

        # Calculate actual dcg:
        top_p = np.argsort(relevance_pred, axis=1)[:, ::-1][:, :k]
        pred_rel = relevance[np.arange(n_instances)[:, None], top_p]
        pred_rel = np.sum(pred_rel / log_term, axis=-1, keepdims=True)
        gain = pred_rel / idcg
        return gain

    return ndcg


def zero_one_rank_loss_for_scores_ties_np(y_true, s_pred):
    n_objects = y_true.shape[1]
    mask = np.greater(y_true[:, None] - y_true[:, :, None], 0).astype(float)
    mask2 = np.greater(s_pred[:, None] - s_pred[:, :, None], 0).astype(float)
    mask3 = np.equal(s_pred[:, None] - s_pred[:, :, None], 0).astype(float)

    # Calculate Transpositions
    transpositions = np.logical_and(mask, mask2)
    x = (np.sum(mask3, axis=(1, 2)) - n_objects).astype(float) / 4.0
    transpositions = np.sum(transpositions, axis=(1, 2)).astype(float)
    transpositions += x

    denominator = n_objects * (n_objects - 1.0) / 2.0
    result = transpositions / denominator
    return np.mean(result)


def zero_one_rank_loss_for_scores_np(y_true, s_pred):
    return zero_one_rank_loss_for_scores_ties_np(y_true, s_pred)


def auc_score(y_true, s_pred):
    idx = np.where(
        (y_true.sum(axis=1) != y_true.shape[-1] - 1)
        & (y_true.sum(axis=1) != y_true.shape[-1])
        & (y_true.sum(axis=1) != 1)
        & (y_true.sum(axis=1) != 0)
    )[0]
    if len(idx) > 1:
        auc = roc_auc_score(y_true[idx], s_pred[idx], average="samples")
    elif len(idx) == 1:
        auc = roc_auc_score(y_true[idx][0], s_pred[idx][0])
    else:
        auc = np.nan
    return auc


def average_precision(y_true, s_pred):
    return average_precision_score(y_true, s_pred, average="samples")


def instance_informedness(y_true, y_pred):
    tp = np.logical_and(y_true, y_pred).sum(axis=1)
    tn = np.logical_and(np.logical_not(y_true), np.logical_not(y_pred)).sum(axis=1)
    cp = y_true.sum(axis=1)
    cn = np.logical_not(y_true).sum(axis=1)
    return np.nanmean(tp / cp + tn / cn - 1)


def f1_measure(y_true, y_pred):
    return f1_score(y_true, y_pred, average="samples")


def precision(y_true, y_pred):
    return precision_score(y_true, y_pred, average="samples")


def recall(y_true, y_pred):
    return recall_score(y_true, y_pred, average="samples")


def subset_01_loss(y_true, y_pred):
    return zero_one_loss(y_true, y_pred)


def hamming(y_true, y_pred):
    return hamming_loss(y_true, y_pred)


def topk_categorical_accuracy_np(k=5):
    def topk_acc(y_true, y_pred):
        topK = y_pred.argsort(axis=1)[:, -k:][:, ::-1]
        accuracies = np.zeros_like(y_true, dtype=bool)
        y_true = np.argmax(y_true, axis=1)
        for i, top in enumerate(topK):
            accuracies[i] = y_true[i] in top
        return np.mean(accuracies)

    return topk_acc


def categorical_accuracy_np(y_true, y_pred):
    y_true = np.argmax(y_true, axis=1)
    choices = np.argmax(y_pred, axis=1)
    accuracies = np.equal(y_true, choices)
    return np.mean(accuracies)


def relevance_gain_np(grading, max_grade):
    """Plain python version of `relevance_gain`, see the documentation
    of that function for details."""
    inverse_grading = -grading + max_grade
    return (2 ** inverse_grading - 1) / (2 ** max_grade)


def err_np(y_true, y_pred, utility_function=None, probability_mapping=None):
    """Plain python version of `err`, see the documentation of that
    function for details.
    """
    if probability_mapping is None:
        # assume y_true is a ranking, use relevance gain
        max_grade = np.max(y_true)
        probability_mapping = partial(relevance_gain_np, max_grade=max_grade)
    if utility_function is None:

        def reciprocal_rank(rank):
            return 1 / rank

        utility_function = reciprocal_rank

    ninstances, nobjects = np.shape(y_pred)

    satisfied_probs = np.reshape(
        list(map(probability_mapping, y_true.flatten())), np.shape(y_true)
    )

    # sort satisfied probabilities according to the predicted ranking
    # black magic invocation to reorder each row
    row_indices = np.arange(ninstances).reshape(-1, 1)
    satisfied_at_rank = satisfied_probs[row_indices, y_pred]

    # still not satisfied after having examined the first r ranks
    not_yet_satisfied_at_rank = np.cumprod(1 - satisfied_at_rank, axis=1)

    # append a column of 1s and drop the last column, since the
    # probability the need has not been satisfied at the 0th rank is 1
    not_yet_satisfied_at_rank = np.hstack(
        (
            np.ones(np.shape(not_yet_satisfied_at_rank)[0]).reshape(-1, 1),
            not_yet_satisfied_at_rank,
        )
    )[:, :-1]

    # map over flattened array, then restore shape
    utilities = list(map(utility_function, range(1, nobjects + 1)))

    discount_at_rank = not_yet_satisfied_at_rank * utilities
    discounted_document_values = satisfied_at_rank * discount_at_rank
    results = np.sum(discounted_document_values, axis=1)
    return np.average(results)



from abc import ABCMeta
from abc import abstractmethod
import inspect
import logging
import os

import numpy as np

from csrank.constants import CHOICE_FUNCTION
from csrank.constants import DISCRETE_CHOICE
from csrank.constants import DYAD_RANKING
from csrank.constants import EXCEPTION_CONTEXT_ARRAY_SHAPE
from csrank.constants import EXCEPTION_OBJECT_ARRAY_SHAPE
from csrank.constants import EXCEPTION_OUTPUT_FEATURES_INSTANCES
from csrank.constants import EXCEPTION_RANKINGS_FEATURES_NO_OF_OBJECTS
from csrank.constants import EXCEPTION_SET_INCLUSION
from csrank.constants import EXCEPTION_UNWANTED_CONTEXT_FEATURES
from csrank.constants import LABEL_RANKING
from csrank.constants import OBJECT_RANKING

logger = logging.getLogger(__name__)


class DatasetReader(metaclass=ABCMeta):
    def __init__(self, dataset_folder="", learning_problem=OBJECT_RANKING, **kwargs):
        """
        The generic dataset parser for parsing datasets for solving different learning problems.

        Parameters
        ----------
        dataset_folder: string
            Folder name in the csank/datasets folder in which
        learning_problem: {'object_ranking', 'dyad_ranking', 'label_ranking', 'discrete_choice', 'choice_function'}
           The learning problem for which the dataset is used.
                - 'object_ranking' learner which will extend :class:`csrank.objectranking.object_ranker.ObjectRanker`
                - 'dyad_ranking' learner which will extend :class:`csrank.dyadranking.dyad_ranker.DyadRanker `
                - 'label_ranking' learner which will extend :class:`csrank.labelranking.label_ranking.LabelRanking`
                - 'discrete_choice' learner which will extend :class:`csrank.discretechoice.discrete_choice.DiscreteObjectChooser`
                - 'choice_function' learner which will extend :class:`csrank.choicefunction.choice_functions.ChoiceFunctions`
        kwargs:
            Keyword arguments for the dataset parser
        """
        logger.info("Learning Problem: {}".format(learning_problem))
        self.X = None
        self.Y = None
        self.Xc = None
        self.learning_problem = learning_problem
        dirname = os.path.dirname(
            os.path.abspath(inspect.getfile(inspect.currentframe()))
        ).replace("dataset_reader", "datasets")
        if dataset_folder is not None:
            self.dirname = os.path.join(dirname, dataset_folder)
            if not os.path.exists(self.dirname):
                logger.warning(
                    "Path given for dataset does not exit {}".format(self.dirname)
                )
        else:
            self.dirname = None

    @abstractmethod
    def __load_dataset__(self):
        raise NotImplementedError

    def __check_dataset_validity__(self):
        if self.X is not None or self.Y is not None:
            if (
                self.learning_problem == OBJECT_RANKING
                or self.learning_problem == DYAD_RANKING
            ):
                assert len(self.X.shape) == 3, EXCEPTION_OBJECT_ARRAY_SHAPE.format(
                    self.learning_problem, self.X.shape
                )
                n_instances, n_objects, n_features = self.X.shape
                assert (
                    n_instances == self.Y.shape[0]
                ), EXCEPTION_OUTPUT_FEATURES_INSTANCES.format(
                    self.learning_problem, self.Y.shape[0], n_instances
                )
                assert (
                    n_objects == self.Y.shape[1]
                ), EXCEPTION_RANKINGS_FEATURES_NO_OF_OBJECTS.format(
                    self.Y.shape[1], n_objects
                )
                if self.learning_problem == OBJECT_RANKING:
                    assert self.Xc is None, EXCEPTION_UNWANTED_CONTEXT_FEATURES.format(
                        self.learning_problem
                    )

            if self.learning_problem == LABEL_RANKING:
                assert len(self.X.shape) == 2, EXCEPTION_CONTEXT_ARRAY_SHAPE.format(
                    self.learning_problem, self.X.shape
                )
                n_instances, n_features = self.X.shape
                assert (
                    n_instances == self.Y.shape[0]
                ), EXCEPTION_OUTPUT_FEATURES_INSTANCES.format(
                    self.learning_problem, self.Y.shape[0], n_instances
                )
                assert self.Xc is None, EXCEPTION_UNWANTED_CONTEXT_FEATURES.format(
                    self.learning_problem
                )

            if self.learning_problem in [DISCRETE_CHOICE, CHOICE_FUNCTION]:
                assert len(self.X.shape) == 3, EXCEPTION_OBJECT_ARRAY_SHAPE.format(
                    self.learning_problem, self.X.shape
                )
                n_instances, n_objects, n_features = self.X.shape
                assert (
                    n_instances == self.Y.shape[0]
                ), EXCEPTION_OUTPUT_FEATURES_INSTANCES.format(
                    self.learning_problem, self.Y.shape[0], n_instances
                )
                assert self.Xc is None, EXCEPTION_UNWANTED_CONTEXT_FEATURES.format(
                    self.learning_problem
                )
                assert n_objects == self.Y.shape[1], EXCEPTION_SET_INCLUSION

            if self.learning_problem == DYAD_RANKING:
                assert len(self.Xc.shape) == 2, EXCEPTION_CONTEXT_ARRAY_SHAPE.format(
                    self.learning_problem, self.Xc.shape
                )
                n_instances, n_features = self.Xc.shape
                assert (
                    n_instances == self.Y.shape[0]
                ), EXCEPTION_OUTPUT_FEATURES_INSTANCES.format(
                    self.learning_problem, self.Y.shape[0], n_instances
                )

    @abstractmethod
    def splitter(self, iter):
        raise NotImplementedError

    @abstractmethod
    def get_dataset_dictionaries(self, lengths=[5, 6]):
        raise NotImplementedError

    @abstractmethod
    def get_single_train_test_split(self):
        raise NotImplementedError

    def get_train_test_datasets(self, n_datasets=5):
        splits = np.array(n_datasets)
        return self.splitter(splits)



import logging

from sklearn.utils import check_random_state

from csrank.learner import Learner
from .object_ranker import ObjectRanker

logger = logging.getLogger(__name__)


class RandomBaselineRanker(ObjectRanker, Learner):
    def __init__(self, random_state=None, **kwargs):
        """
        Baseline assigns the average number of chosen objects in the given choice sets and chooses all the objects.

        :param kwargs: Keyword arguments for the algorithms
        """

        self.random_state = random_state

    def _pre_fit(self):
        super()._pre_fit()
        self.random_state_ = check_random_state(self.random_state)

    def fit(self, X, Y, **kwd):
        self._pre_fit()
        return self

    def _predict_scores_fixed(self, X, **kwargs):
        n_instances, n_objects, n_features = X.shape
        return self.random_state_.rand(n_instances, n_objects)



import logging

from sklearn.model_selection import train_test_split

from csrank.core.pairwise_svm import PairwiseSVM
from .choice_functions import ChoiceFunctions
from .util import generate_complete_pairwise_dataset

logger = logging.getLogger(__name__)


class PairwiseSVMChoiceFunction(ChoiceFunctions, PairwiseSVM):
    def __init__(
        self,
        C=1.0,
        tol=1e-4,
        normalize=True,
        fit_intercept=True,
        random_state=None,
        **kwargs,
    ):
        """
        Create an instance of the :class:`PairwiseSVM` model for learning a choice function.
        It learns a linear deterministic utility function of the form :math:`U(x) = w \\cdot x`, where :math:`w` is
        the weight vector. It is estimated using *pairwise preferences* generated from the choices.
        The choice set is defined as:

        .. math::

            c(Q) = \\{ x_i \\in Q \\lvert \\, U(x_i) > t \\}

        Parameters
        ----------
        C : float, optional
            Penalty parameter of the error term
        tol : float, optional
            Optimization tolerance
        normalize : bool, optional
            If True, the data will be normalized before fitting.
        fit_intercept : bool, optional
            If True, the linear model will also fit an intercept.
        random_state : int, RandomState instance or None, optional
            Seed of the pseudorandom generator or a RandomState instance
        **kwargs
            Keyword arguments for the algorithms


        References
        ----------
            [1] Theodoros Evgeniou, Massimiliano Pontil, and Olivier Toubia.„A convex optimization approach to modeling consumer heterogeneity in conjoint estimation“. In: Marketing Science 26.6 (2007), pp. 805–818.

            [2] Sebastián Maldonado, Ricardo Montoya, and Richard Weber. „Advanced conjoint analysis using feature selection via support vector machines“. In: European Journal of Operational Research 241.2 (2015), pp. 564 –574.

        """
        super().__init__(
            C=C,
            tol=tol,
            normalize=normalize,
            fit_intercept=fit_intercept,
            random_state=random_state,
            **kwargs,
        )

    def _convert_instances_(self, X, Y):
        logger.debug("Creating the Dataset")
        (
            garbage,
            garbage,
            x_train,
            garbage,
            y_single,
        ) = generate_complete_pairwise_dataset(X, Y)
        del garbage
        assert x_train.shape[1] == self.n_object_features_fit_
        logger.debug("Finished the Dataset with instances {}".format(x_train.shape[0]))
        return x_train, y_single

    def fit(self, X, Y, tune_size=0.1, thin_thresholds=1, verbose=0, **kwd):
        """
        Fit a generic preference learning model on a provided set of queries.
        The provided queries can be of a fixed size (numpy arrays).

        Parameters
        ----------
        X : numpy array (n_instances, n_objects, n_features)
            Feature vectors of the objects
        Y : numpy array (n_instances, n_objects)
            Choices for given objects in the query
        tune_size: float (range : [0,1])
            Percentage of instances to split off to tune the threshold for the choice function
        thin_thresholds: int
            The number of instances of scores to skip while tuning the threshold
        verbose : bool
            Print verbose information
        **kwd :
            Keyword arguments for the fit function

        """
        self._pre_fit()
        _n_instances, self.n_objects_fit_, self.n_object_features_fit_ = X.shape
        if tune_size > 0:
            X_train, X_val, Y_train, Y_val = train_test_split(
                X, Y, test_size=tune_size, random_state=self.random_state
            )
            try:
                super().fit(X_train, Y_train, **kwd)
            finally:
                logger.info(
                    "Fitting utility function finished. Start tuning threshold."
                )
                self.threshold_ = self._tune_threshold(
                    X_val, Y_val, thin_thresholds=thin_thresholds, verbose=verbose
                )
        else:
            super().fit(X, Y, **kwd)
            self.threshold_ = 0.5
        return self



from abc import ABCMeta
import collections
import logging
import os

import numpy as np
from pandas.api.types import is_numeric_dtype
from sklearn.model_selection import ShuffleSplit

from csrank.dataset_reader.dataset_reader import DatasetReader
from csrank.dataset_reader.util import standardize_features
from csrank.util import print_dictionary

try:
    import h5py
except ImportError:
    from csrank.util import MissingExtraError

    raise MissingExtraError("h5py", "data")

try:
    import pandas as pd
except ImportError:
    from csrank.util import MissingExtraError

    raise MissingExtraError("pandas", "data")

logger = logging.getLogger(__name__)


class ExpediaDatasetReader(DatasetReader, metaclass=ABCMeta):
    def __init__(self, fold_id=0, **kwargs):
        super(ExpediaDatasetReader, self).__init__(dataset_folder="expedia", **kwargs)
        self.RAW_DATASET_FILE = os.path.join(self.dirname, "train.csv")
        self.fold_id = fold_id
        self.X_train = self.Y_train = self.X_test = self.Y_test = dict()

    def __load_dataset__(self):
        if not os.path.isfile(self.hdf5file_path):
            self.create_choices_dataset()
        self.X_dict, self.Y_dict = self.get_choices_dict()
        logger.info("Done loading the dataset")

    def get_choices_dict(self):
        file = h5py.File(self.hdf5file_path, "r")
        lengths = file["lengths"]
        X = dict()
        Y = dict()
        for ranking_length in np.array(lengths):
            self.X = np.array(file["X_{}".format(ranking_length)])
            self.Y = np.array(file["Y_{}".format(ranking_length)])
            self.__check_dataset_validity__()
            X[ranking_length], Y[ranking_length] = self.X, self.Y
        file.close()
        return X, Y

    def create_choices_dataset(self):
        logger.info("Writing in hd5 {}".format(self.hdf5file_path))
        X, scores = self._parse_dataset()
        result, freq = self._build_training_buckets(X, scores)
        h5f = h5py.File(self.hdf5file_path, "w")
        logger.info("Frequencies of rankings: {}".format(print_dictionary(freq)))

        for key, value in result.items():
            x, s = value
            h5f.create_dataset(
                "X_" + str(key), data=x, compression="gzip", compression_opts=9
            )
            h5f.create_dataset(
                "Y_" + str(key), data=s, compression="gzip", compression_opts=9
            )
        lengths = np.sort(np.array(list(result.keys())))
        h5f.create_dataset(
            "lengths", data=lengths, compression="gzip", compression_opts=9
        )
        h5f.close()

    def _build_training_buckets(self, X, Y):
        """Separates object ranking data into buckets of the same ranking size."""
        result = dict()
        frequencies = dict()

        for x, y in zip(X, Y):
            n_objects = len(x)
            if n_objects not in result:
                result[n_objects] = ([], [])
            bucket = result[n_objects]
            bucket[0].append(x)
            bucket[1].append(y)
            if n_objects not in frequencies:
                frequencies[n_objects] = 1
            else:
                frequencies[n_objects] += 1

        # Convert all buckets to numpy arrays:
        for k, v in result.items():
            result[k] = np.array(v[0]), np.array(v[1])
        result = collections.OrderedDict(sorted(result.items()))
        return result, frequencies

    def _filter_dataset(self):
        train = pd.read_csv(self.RAW_DATASET_FILE)
        for col in train.columns.values:
            if "id" in col and col != "srch_id":
                del train[col]
            elif is_numeric_dtype(train[col]):
                arr = np.array(train[col])
                fraction = np.isnan(arr).sum() / len(arr)
                if fraction > 0.0:
                    logger.info(
                        "########################################################################"
                    )
                    logger.info(
                        "Missing values {}: {}".format(
                            col, np.isnan(arr).sum() / len(arr)
                        )
                    )
                    logger.info("Min {}: Max {}".format(np.nanmin(arr), np.nanmax(arr)))
                    if fraction > 0.50:
                        del train[col]
                    else:
                        train.loc[train[col].isnull(), col] = np.nanmin(arr) - 1
            else:
                del train[col]
        self.train_df = train

    def sub_sampling_from_dictionary(self):
        X = []
        Y = []
        for n in self.X_train.keys():
            if n == self.n_objects:
                x, y = self.X_train[n], self.Y_train[n]
            elif n > self.n_objects:
                x, y = self.sub_sampling_function(self.X_train[n], self.Y_train[n])
            else:
                x = []
            if len(x) != 0:
                if len(X) == 0:
                    X = np.copy(x)
                    Y = np.copy(y)
                else:
                    X = np.concatenate([X, x], axis=0)
                    Y = np.concatenate([Y, y], axis=0)
                logger.info(
                    "Sampled instances {} from objects {}".format(x.shape[0], n)
                )
        logger.info("Sampled instances {} objects {}".format(X.shape[0], X.shape[1]))
        return X, Y

    def splitter(self, iter):
        for train_idx, test_idx in iter:
            x_train, y_train, x_test, y_test = (
                self.X[train_idx],
                self.Y[train_idx],
                self.X[test_idx],
                self.Y[test_idx],
            )
            x_train, x_test = standardize_features(x_train, x_test)
            yield x_train, y_train, x_test, y_test

    def get_single_train_test_split(self):
        splits = dict()
        cv_iter = ShuffleSplit(
            n_splits=1, random_state=self.random_state, test_size=0.80
        )
        for n_obj, arr in self.X_dict.items():
            if arr.shape[0] == 1:
                splits[n_obj] = ([0], [0])
            else:
                splits[n_obj] = list(cv_iter.split(arr))[0]
        self.X_train = dict()
        self.Y_train = dict()
        self.X_test = dict()
        self.Y_test = dict()
        for n_obj, itr in splits.items():
            train_idx, test_idx = itr
            self.X_train[n_obj] = np.copy(self.X_dict[n_obj][train_idx])
            self.X_test[n_obj] = np.copy(self.X_dict[n_obj][test_idx])
            self.Y_train[n_obj] = np.copy(self.Y_dict[n_obj][train_idx])
            self.Y_test[n_obj] = np.copy(self.Y_dict[n_obj][test_idx])
        self.X, self.Y = self.sub_sampling_from_dictionary()
        self.__check_dataset_validity__()
        self.X, self.X_test = standardize_features(self.X, self.X_test)
        return self.X, self.Y, self.X_test, self.Y_test

    def sub_sampling_function(self, Xt, Yt):
        raise NotImplementedError

    def _parse_dataset(self):
        raise NotImplementedError

    def get_dataset_dictionaries(self):
        self.X_test, self.X_test = standardize_features(self.X, self.X_test)
        return self.X_train, self.Y_train, self.X_test, self.Y_test



# We should re-evaluate if we really want to re-export everything here and then
# use __all__ properly.

from csrank.dataset_reader.choicefunctions import *  # noqa
from csrank.dataset_reader.discretechoice import *  # noqa
from csrank.dataset_reader.dyadranking import *  # noqa
from csrank.dataset_reader.expedia_dataset_reader import ExpediaDatasetReader  # noqa
from csrank.dataset_reader.labelranking import *  # noqa
from csrank.dataset_reader.objectranking import *  # noqa
from csrank.dataset_reader.synthetic_dataset_generator import SyntheticIterator  # noqa



import logging

import numpy as np
from sklearn.cluster import MiniBatchKMeans
from sklearn.utils import check_random_state

from csrank.discretechoice.likelihoods import create_weight_dictionary
from csrank.discretechoice.likelihoods import fit_pymc3_model
from csrank.learner import Learner
import csrank.numpy_util as npu
import csrank.theano_util as ttu
from csrank.util import print_dictionary
from .discrete_choice import DiscreteObjectChooser
from .likelihoods import likelihood_dict
from .likelihoods import LogLikelihood

try:
    import pymc3 as pm
    from pymc3.variational.callbacks import CheckParametersConvergence
except ImportError:
    from csrank.util import MissingExtraError

    raise MissingExtraError("pymc3", "probabilistic")

try:
    import theano
    from theano import tensor as tt
except ImportError:
    from csrank.util import MissingExtraError

    raise MissingExtraError("theano", "probabilistic")

logger = logging.getLogger(__name__)


class NestedLogitModel(DiscreteObjectChooser, Learner):
    def __init__(
        self,
        n_nests=None,
        loss_function="",
        regularization="l1",
        alpha=1e-2,
        random_state=None,
        **kwd,
    ):
        """
        Create an instance of the Nested Logit model for learning the discrete choice function. This model divides
        objects into disjoint subsets called nests,such that the objects which are similar to each other are in same
        nest. This model structure is 1-layer of hierarchy and the :math:`\\lambda` for each nest :math:`B_k` signifies
        the degree of independence and  :math:`1-\\lambda` signifies the correlations between the object in it. We
        learn two weight vectors and the  :math:`\\lambda s`.

        The probability of choosing an object :math:`x_i` from the given query set :math:`Q` is defined by product
        of choosing the nest in which :math:`x_i` exists and then choosing the the object from the nest.

        .. math::

            P(x_i \\lvert Q) = P_i = P_{i \\lvert B_k} P_{B_k} \\enspace ,


        The discrete choice for the given query set :math:`Q` is defined as:

        .. math::

            dc(Q) := \\operatorname{argmax}_{x_i \\in Q }  \\; P(x_i \\lvert Q)

        Parameters
        ----------
        n_nests : int range : [2,n_objects/2]
            The number of nests/subsets in which the objects are divided.
            This may not surpass half the amount of objects this model will
            be trained on.
        loss_function : string , {‘categorical_crossentropy’, ‘binary_crossentropy’, ’categorical_hinge’}
            Loss function to be used for the discrete choice decision from the query set
        regularization : string, {‘l1’, ‘l2’}, string
           Regularizer function (L1 or L2) applied to the `kernel` weights matrix
        alpha: float (range : [0,1])
            The lower bound of the correlations between the objects in a nest
        random_state : int or object
            Numpy random state
        **kwargs
            Keyword arguments for the algorithms

        References
        ----------
            [1] Kenneth E Train. „Discrete choice methods with simulation“. In: Cambridge university press, 2009. Chap GEV, pp. 87–111.

            [2] Kenneth Train. Qualitative choice analysis. Cambridge, MA: MIT Press, 1986

            [3] Kenneth Train and Daniel McFadden. „The goods/leisure tradeoff and disaggregate work trip mode choice models“. In: Transportation research 12.5 (1978), pp. 349–353
        """
        self.n_nests = n_nests
        self.alpha = alpha
        self.random_state = random_state
        self.loss_function = loss_function
        known_regularization_functions = {"l1", "l2"}
        if regularization not in known_regularization_functions:
            raise ValueError(
                f"Regularization function {regularization} is unknown. Must be one of {known_regularization_functions}"
            )
        self.regularization = regularization

    @property
    def model_configuration(self):
        """
            Constructs the dictionary containing the priors for the weight vectors for the model according to the
            regularization function. The parameters are:
                * **weights** : Weights to evaluates the utility of the objects
                * **weights_k** : Weights to evaluates the utility of the nests

            For ``l1`` regularization the priors are:

            .. math::

                \\text{mu}_w \\sim \\text{Normal}(\\text{mu}=0, \\text{sd}=5.0) \\\\
                \\text{b}_w \\sim \\text{HalfCauchy}(\\beta=1.0) \\\\
                \\text{weights} \\sim \\text{Laplace}(\\text{mu}=\\text{mu}_w, \\text{b}=\\text{b}_w)

            For ``l2`` regularization the priors are:

            .. math::

                \\text{mu}_w \\sim \\text{Normal}(\\text{mu}=0, \\text{sd}=5.0) \\\\
                \\text{sd}_w \\sim \\text{HalfCauchy}(\\beta=1.0) \\\\
                \\text{weights} \\sim \\text{Normal}(\\text{mu}=\\text{mu}_w, \\text{sd}=\\text{sd}_w)


            Returns
            -------
                configuration : dict
                    Dictionary containing the priors applies on the weights
        """
        if not hasattr(self, "config_"):
            if self.regularization == "l2":
                weight = pm.Normal
                prior = "sd"
            elif self.regularization == "l1":
                weight = pm.Laplace
                prior = "b"
            self.config_ = {
                "weights": [
                    weight,
                    {
                        "mu": (pm.Normal, {"mu": 0, "sd": 5}),
                        prior: (pm.HalfCauchy, {"beta": 1}),
                    },
                ],
                "weights_k": [
                    weight,
                    {
                        "mu": (pm.Normal, {"mu": 0, "sd": 5}),
                        prior: (pm.HalfCauchy, {"beta": 1}),
                    },
                ],
            }
            logger.info(
                "Creating model with config {}".format(print_dictionary(self.config_))
            )
        return self.config_

    def create_nests(self, X):
        """
        For allocating the objects to different nests we use the clustering algorithm with number of clusters
        :math:`k` and allocate the similar objects in query set :math:`Q`.

        Parameters
        ----------
        X : numpy array
            (n_instances, n_objects, n_features)
            Feature vectors of the objects in the query sets

        Returns
        -------
        Yn : numpy array
            (n_instances, n_objects) Values for each object implying the nest it belongs to. For example for :math:`2` nests the value 0 implies that object is allocated to nest 1 and value 1 implies it is allocated to nest 2.

        """
        self.random_state_ = self.random_state_
        n, n_obj, n_dim = X.shape
        objects = X.reshape(n * n_obj, n_dim)
        if not hasattr(self, "cluster_model_"):
            self.cluster_model_ = MiniBatchKMeans(
                n_clusters=self.n_nests, random_state=self.random_state_
            ).fit(objects)
            self.features_nests_ = self.cluster_model_.cluster_centers_
            prediction = self.cluster_model_.labels_
        else:
            prediction = self.cluster_model_.predict(objects)
        Yn = []
        for i in np.arange(0, n * n_obj, step=n_obj):
            nest_ids = prediction[i : i + n_obj]
            Yn.append(nest_ids)
        Yn = np.array(Yn)
        return Yn

    def _eval_utility(self, weights):
        utility = tt.zeros(tuple(self.y_nests_.shape))
        for i in range(self.n_nests):
            rows, cols = tt.eq(self.y_nests_, i).nonzero()
            utility = tt.set_subtensor(
                utility[rows, cols], tt.dot(self.Xt_[rows, cols], weights[i])
            )
        return utility

    def get_probabilities(self, utility, lambda_k, utility_k):
        """
            This method calculates the probability of choosing an object from the query set using the following parameters of the model which are used:

                * **weights** (:math:`w`): Weights to get the utility of the object :math:`Y_i = U(x_i) = w \\cdot x_i`
                * **weights_k** (:math:`w_k`): Weights to get the utility of the next  :math:`W_k = U_k(x) = w_k \\cdot c_k`, where :math:`c_k` is the center of the object space of nest :math:`B_k`
                * **lambda_k** (:math:`\\lambda_k`): Lambda is the measure of independence amongst the obejcts in the nest :math:`B_k`

            The probability of choosing the object  :math:`x_i` from the query set :math:`Q`:

            .. math::
                P_i = \\frac{\\boldsymbol{e}^{ ^{Y_i} /_{\\lambda_k}}}{\\sum_{j \\in B_k} \\boldsymbol{e}^{^{Y_j} /_{\\lambda_k}}} \\frac {\\boldsymbol{e}^{W_k + \\lambda_k I_k}} {\\sum_{\\ell = 1}^{K} \\boldsymbol{e}^{ W_{\\ell } + \\lambda_{\\ell} I_{\\ell}}} \\quad i \\in B_k  \\enspace , \\\\
                where,\\enspace I_k = \\ln \\sum_{ j \\in B_k} \\boldsymbol{e}^{^{Y_j} /_{\\lambda_k}}


            Parameters
            ----------
            utility : theano tensor
                (n_instances, n_objects)
                Utility :math:`Y_i` of the objects :math:`x_i \\in Q` in the query sets
            lambda_k : theano tensor (range : [alpha, 1.0])
                (n_nests)
                Measure of independence amongst the obejcts in each nests
            utility_k : theano tensor
                (n_instances, n_nests)
                Utilities of the nests :math:`B_k \\in \\mathcal{B}`

            Returns
            -------
            p : theano tensor
                (n_instances, n_objects)
                Choice probabilities :math:`P_i` of the objects :math:`x_i \\in Q` in the query sets

        """
        n_instances, n_objects = self.y_nests_.shape
        pni_k = tt.zeros((n_instances, n_objects))
        ivm = tt.zeros((n_instances, self.n_nests))
        for i in range(self.n_nests):
            rows, cols = tt.neq(self.y_nests_, i).nonzero()
            sub_tensor = tt.set_subtensor(utility[rows, cols], -1e50)
            ink = ttu.logsumexp(sub_tensor)
            rows, cols = tt.eq(self.y_nests_, i).nonzero()
            pni_k = tt.set_subtensor(
                pni_k[rows, cols], tt.exp(sub_tensor - ink)[rows, cols]
            )
            ivm = tt.set_subtensor(ivm[:, i], lambda_k[i] * ink[:, 0] + utility_k[i])
        pk = tt.exp(ivm - ttu.logsumexp(ivm))
        pn_k = tt.zeros((n_instances, n_objects))
        for i in range(self.n_nests):
            rows, cols = tt.eq(self.y_nests_, i).nonzero()
            p = tt.ones((n_instances, n_objects)) * pk[:, i][:, None]
            pn_k = tt.set_subtensor(pn_k[rows, cols], p[rows, cols])
        p = pni_k * pn_k
        return p

    def _eval_utility_np(self, x_t, y_nests, weights):
        utility = np.zeros(tuple(y_nests.shape))
        for i in range(self.n_nests):
            rows, cols = np.where(y_nests == i)
            utility[rows, cols] = np.dot(x_t[rows, cols], weights[i])
        return utility

    def _get_probabilities_np(self, Y_n, utility, lambda_k, utility_k):
        n_instances, n_objects = Y_n.shape
        pni_k = np.zeros((n_instances, n_objects))
        ivm = np.zeros((n_instances, self.n_nests))
        for i in range(self.n_nests):
            sub_tensor = np.copy(utility)
            sub_tensor[np.where(Y_n != i)] = -1e50
            ink = npu.logsumexp(sub_tensor)
            pni_k[np.where(Y_n == i)] = np.exp(sub_tensor - ink)[np.where(Y_n == i)]
            ivm[:, i] = lambda_k[i] * ink[:, 0] + utility_k[i]
        pk = np.exp(ivm - npu.logsumexp(ivm))
        pn_k = np.zeros((n_instances, n_objects))
        for i in range(self.n_nests):
            rows, cols = np.where(Y_n == i)
            p = np.ones((n_instances, n_objects)) * pk[:, i][:, None]
            pn_k[rows, cols] = p[rows, cols]
        p = pni_k * pn_k
        return p

    def construct_model(self, X, Y):
        """
        Constructs the nested logit model by applying priors on weight vectors **weights** and **weights_k** as per
        :meth:`model_configuration`. Then we apply a uniform prior to the :math:`\\lambda s`, i.e.
        :math:`\\lambda s \\sim Uniform(\\text{alpha}, 1.0)`.The probability of choosing the object :math:`x_i` from
        the query set :math:`Q = \\{x_1, \\ldots ,x_n\\}` is evaluated in :meth:`get_probabilities`.

        Parameters
        ----------
        X : numpy array
            (n_instances, n_objects, n_features)
            Feature vectors of the objects
        Y : numpy array
            (n_instances, n_objects)
            Preferences in the form of discrete choices for given objects

        Returns
        -------
         model : pymc3 Model :class:`pm.Model`
        """
        self.loss_function_ = likelihood_dict.get(self.loss_function, None)
        self.threshold_ = 5e6
        self.trace_ = None
        self.trace_vi_ = None
        if np.prod(X.shape) > self.threshold_:
            upper_bound = int(self.threshold_ / np.prod(X.shape[1:]))
            indices = self.random_state_.choice(X.shape[0], upper_bound, replace=False)
            X = X[indices, :, :]
            Y = Y[indices, :]
        logger.info("Train Set instances {} objects {} features {}".format(*X.shape))
        y_nests = self.create_nests(X)
        with pm.Model() as self.model:
            self.Xt_ = theano.shared(X)
            self.Yt_ = theano.shared(Y)
            self.y_nests_ = theano.shared(y_nests)
            shapes = {
                "weights": self.n_object_features_fit_,
                "weights_k": self.n_object_features_fit_,
            }

            weights_dict = create_weight_dictionary(self.model_configuration, shapes)
            lambda_k = pm.Uniform("lambda_k", self.alpha, 1.0, shape=self.n_nests)
            weights = weights_dict["weights"] / lambda_k[:, None]
            utility = self._eval_utility(weights)
            utility_k = tt.dot(self.features_nests_, weights_dict["weights_k"])
            self.p_ = self.get_probabilities(utility, lambda_k, utility_k)

            LogLikelihood(
                "yl", loss_func=self.loss_function_, p=self.p_, observed=self.Yt_
            )
        logger.info("Model construction completed")

    def fit(
        self,
        X,
        Y,
        sampler="variational",
        tune=500,
        draws=500,
        vi_params={
            "n": 20000,
            "method": "advi",
            "callbacks": [CheckParametersConvergence()],
        },
        **kwargs,
    ):
        """
        Fit a nested logit model on the provided set of queries X and choices Y of those objects. The provided
        queries and corresponding preferences are of a fixed size (numpy arrays). For learning this network the
        categorical cross entropy loss function for each object :math:`x_i \\in Q` is defined as:

        .. math::

            C_{i} =  -y(i)\\log(P_i) \\enspace,

        where :math:`y` is ground-truth discrete choice vector of the objects in the given query set :math:`Q`.
        The value :math:`y(i) = 1` if object :math:`x_i` is chosen else :math:`y(i) = 0`.

        Parameters
        ----------
        X : numpy array (n_instances, n_objects, n_features)
            Feature vectors of the objects
        Y : numpy array (n_instances, n_objects)
            Choices for given objects in the query
        sampler : {‘variational’, ‘metropolis’, ‘nuts’}, string
            The sampler used to estimate the posterior mean and mass matrix from the trace

                * **variational** : Run inference methods to estimate posterior mean and diagonal mass matrix
                * **metropolis** : Use the MAP as starting point and Metropolis-Hastings sampler
                * **nuts** : Use the No-U-Turn sampler
        vi_params : dict
            The parameters for the **variational** inference method
        draws : int
            The number of samples to draw. Defaults to 500. The number of tuned samples are discarded by default
        tune : int
            Number of iterations to tune, defaults to 500. Ignored when using 'SMC'. Samplers adjust
            the step sizes, scalings or similar during tuning. Tuning samples will be drawn in addition
            to the number specified in the `draws` argument, and will be discarded unless
            `discard_tuned_samples` is set to False.
        **kwargs :
            Keyword arguments for the fit function of :meth:`pymc3.fit`or :meth:`pymc3.sample`
        """
        self._pre_fit()
        _n_instances, self.n_objects_fit_, self.n_object_features_fit_ = X.shape
        if self.n_nests is None:
            self.n_nests = int(self.n_objects_fit_ / 2)
        self.random_state_ = check_random_state(self.random_state)
        self.construct_model(X, Y)
        fit_pymc3_model(self, sampler, draws, tune, vi_params, **kwargs)
        return self

    def _predict_scores_fixed(self, X, **kwargs):
        y_nests = self.create_nests(X)
        mean_trace = dict(pm.summary(self.trace_)["mean"])
        weights = np.array(
            [
                mean_trace["weights[{}]".format(i)]
                for i in range(self.n_object_features_fit_)
            ]
        )
        weights_k = np.array(
            [
                mean_trace["weights_k[{}]".format(i)]
                for i in range(self.n_object_features_fit_)
            ]
        )
        lambda_k = np.array(
            [mean_trace["lambda_k[{}]".format(i)] for i in range(self.n_nests)]
        )
        weights = weights / lambda_k[:, None]
        utility_k = np.dot(self.features_nests_, weights_k)
        utility = self._eval_utility_np(X, y_nests, weights)
        scores = self._get_probabilities_np(y_nests, utility, lambda_k, utility_k)
        return scores



import functools

import torch.nn as nn

from csrank.discrete_choice_losses import CategoricalHingeLossMax
from csrank.discretechoice.discrete_choice import SkorchDiscreteChoiceFunction
from csrank.modules.object_mapping import DenseNeuralNetwork
from csrank.modules.scoring import FATEScoring


class FATEDiscreteChoiceFunction(SkorchDiscreteChoiceFunction):
    """A discrete choice estimator based on the FATE-Approach.

    See the documentation of :class:`csrank.modules.scoring.FATEScoring` for
    more details.

    Parameters
    ----------
    n_hidden_set_layers : int
        The number of hidden layers that should be used for the ``DeepSet``
        context embedding.

    n_hidden_set_untis : int
        The number of units per hidden layer that should be used for the
        ``DeepSet`` context embedding.

    n_hidden_joint_layers : int
        The number of hidden layers that should be used for the utility
        function that evaluates each object in the aggregated context.

    n_hidden_joint_units : int
        The number of units per hidden layer that should used for the utility
        function that evaluates each object in the aggregated context.

    activation : torch activation function (class)
        The activation function that should be used for each layer of the two
        ("set" and "joint) neural networks.

    choice_size : int
        The size of the target choice set.

    criterion : torch criterion (class)
        The criterion that is used to evaluate and optimize the module.

    **kwargs : skorch NeuralNet arguments
        All keyword arguments are passed to the constructor of
        ``SkorchDiscreteChoice``. See the documentation of that class for more
        details.
    """

    def __init__(
        self,
        n_hidden_set_layers=2,
        n_hidden_set_units=32,
        n_hidden_joint_layers=2,
        n_hidden_joint_units=32,
        activation=nn.SELU,
        choice_size=1,
        criterion=CategoricalHingeLossMax,
        **kwargs
    ):
        self.n_hidden_set_layers = n_hidden_set_layers
        self.n_hidden_set_units = n_hidden_set_units
        self.n_hidden_joint_layers = n_hidden_joint_layers
        self.n_hidden_joint_units = n_hidden_joint_units
        self.activation = activation
        super().__init__(
            module=FATEScoring, criterion=criterion, choice_size=choice_size, **kwargs
        )

    def _get_extra_module_parameters(self):
        """Return extra parameters that should be passed to the module."""
        params = super()._get_extra_module_parameters()
        params["pairwise_utility_module"] = functools.partial(
            DenseNeuralNetwork,
            hidden_layers=self.n_hidden_joint_layers,
            units_per_hidden=self.n_hidden_joint_units,
            activation=self.activation(),
            output_size=1,
        )
        params["embedding_module"] = functools.partial(
            DenseNeuralNetwork,
            hidden_layers=self.n_hidden_set_layers,
            units_per_hidden=self.n_hidden_set_units,
            activation=self.activation(),
        )
        return params



from .baseline import RandomBaselineRanker
from .cmpnet_object_ranker import CmpNetObjectRanker
from .expected_rank_regression import ExpectedRankRegression
from .fate_object_ranker import FATEObjectRanker
from .feta_object_ranker import FETAObjectRanker
from .rank_svm import RankSVM
from .ranknet_object_ranker import RankNetObjectRanker

__all__ = [
    "CmpNetObjectRanker",
    "ExpectedRankRegression",
    "FATEObjectRanker",
    "FETAObjectRanker",
    "RandomBaselineRanker",
    "RankNetObjectRanker",
    "RankSVM",
]



import logging

import numpy as np
from sklearn.linear_model import LogisticRegression
from sklearn.preprocessing import StandardScaler
from sklearn.svm import LinearSVC
from sklearn.utils import check_random_state

from csrank.learner import Learner

logger = logging.getLogger(__name__)


class PairwiseSVM(Learner):
    def __init__(
        self,
        C=1.0,
        tol=1e-4,
        normalize=True,
        fit_intercept=True,
        use_logistic_regression=False,
        random_state=None,
        **kwargs,
    ):
        """Create an instance of the PairwiseSVM model for any preference learner.

        Parameters
        ----------
        C : float, optional
            Penalty parameter of the error term
        tol : float, optional
            Optimization tolerance
        normalize : bool, optional
            If True, the data will be normalized before fitting.
        fit_intercept : bool, optional
            If True, the linear model will also fit an intercept.
        use_logistic_regression : bool, optional
            Whether to fit a Linear Support Vector machine or a Logistic
            Regression model. You may want to prefer the simpler Logistic
            Regression model on a large sample size.
        random_state : int, RandomState instance or None, optional
            Seed of the pseudorandom generator or a RandomState instance
        **kwargs
            Keyword arguments for the algorithms

        References
        ----------
            [1] Joachims, T. (2002, July). "Optimizing search engines using clickthrough data.", Proceedings of the eighth ACM SIGKDD international conference on Knowledge discovery and data mining (pp. 133-142). ACM.
        """
        self.normalize = normalize
        self.C = C
        self.tol = tol
        self.use_logistic_regression = use_logistic_regression
        self.random_state = random_state
        self.fit_intercept = fit_intercept

    def _pre_fit(self):
        super()._pre_fit()
        self.random_state_ = check_random_state(self.random_state)

    def fit(self, X, Y, **kwargs):
        """
        Fit a generic preference learning model on a provided set of queries.
        The provided queries can be of a fixed size (numpy arrays).

        Parameters
        ----------
        X : numpy array, shape (n_samples, n_objects, n_features)
            Feature vectors of the objects
        Y : numpy array, shape (n_samples, n_objects, n_features)
            Preferences in form of Orderings or Choices for given n_objects
        **kwargs
            Keyword arguments for the fit function

        """
        self._pre_fit()
        _n_instances, self.n_objects_fit_, self.n_object_features_fit_ = X.shape
        if self.use_logistic_regression:
            self.model_ = LogisticRegression(
                C=self.C,
                tol=self.tol,
                fit_intercept=self.fit_intercept,
                random_state=self.random_state_,
            )
            logger.info("Logistic Regression model ")
        else:
            self.model_ = LinearSVC(
                C=self.C,
                tol=self.tol,
                fit_intercept=self.fit_intercept,
                random_state=self.random_state_,
            )
            logger.info("Linear SVC model ")

        if self.n_objects_fit_ < 2:
            # Nothing to learn, cannot create pairwise instances.
            return self
        x_train, y_single = self._convert_instances_(X, Y)
        if self.normalize:
            std_scalar = StandardScaler()
            x_train = std_scalar.fit_transform(x_train)
        logger.debug("Finished Creating the model, now fitting started")

        self.model_.fit(x_train, y_single)
        self.weights_ = self.model_.coef_.flatten()
        if self.fit_intercept:
            self.weights_ = np.append(self.weights_, self.model_.intercept_)
        logger.debug("Fitting Complete")
        return self

    def _predict_scores_fixed(self, X, **kwargs):
        assert X.shape[-1] == self.n_object_features_fit_
        logger.info("For Test instances {} objects {} features {}".format(*X.shape))
        if self.fit_intercept:
            scores = np.dot(X, self.weights_[:-1])
        else:
            scores = np.dot(X, self.weights_)
        logger.info("Done predicting scores")
        return np.array(scores)

    def _convert_instances_(self, X, Y):
        raise NotImplementedError



import numpy as np
import pytest
import torch
from torch import optim

from csrank.constants import CMPNET
from csrank.constants import ERR
from csrank.constants import FATE_RANKER
from csrank.constants import FETA_RANKER
from csrank.constants import RANKNET
from csrank.constants import RANKSVM
from csrank.metrics_np import zero_one_accuracy_np
from csrank.metrics_np import zero_one_rank_loss_for_scores_ties_np
from csrank.objectranking import CmpNetObjectRanker
from csrank.objectranking import ExpectedRankRegression
from csrank.objectranking import FATEObjectRanker
from csrank.objectranking import FETAObjectRanker
from csrank.objectranking import RankNetObjectRanker
from csrank.objectranking import RankSVM

skorch_common_args = {
    "max_epochs": 100,
    "optimizer": optim.SGD,
    "optimizer__lr": 1e-3,
    "optimizer__momentum": 0.9,
    "optimizer__nesterov": True,
    # We evaluate the estimators in-sample. These tests are just small
    # sanity checks, so overfitting is okay here.
    "train_split": None,
}

object_rankers = {
    CMPNET: (
        CmpNetObjectRanker,
        {"n_hidden": 2, "n_units": 8, **skorch_common_args},
        (0.0, 1.0),
    ),
    ERR: (ExpectedRankRegression, {}, (0.0, 1.0)),
    RANKSVM: (RankSVM, {}, (0.0, 1.0)),
    FATE_RANKER: (
        FATEObjectRanker,
        {
            "n_hidden_joint_layers": 1,
            "n_hidden_set_layers": 1,
            "n_hidden_joint_units": 5,
            "n_hidden_set_units": 5,
            **skorch_common_args,
        },
        (0.0, 1.0),
    ),
    FETA_RANKER: (
        FETAObjectRanker,
        {
            "n_hidden": 1,
            "n_units": 8,
            "add_zeroth_order_model": False,
            **skorch_common_args,
        },
        (0.0, 1.0),
    ),
    FETA_RANKER
    + "zeroth_order_model": (
        FETAObjectRanker,
        {
            "n_hidden": 1,
            "n_units": 8,
            "add_zeroth_order_model": True,
            **skorch_common_args,
        },
        (0.0, 1.0),
    ),
    RANKNET: (
        RankNetObjectRanker,
        {"n_hidden": 2, "n_units": 8, **skorch_common_args},
        (0.0, 1.0),
    ),
}


@pytest.fixture(scope="module")
def trivial_ranking_problem():
    random_state = np.random.RandomState(123)
    # pytorch uses 32 bit floats by default. That should be precise enough and
    # makes it easier to use pytorch and non-pytorch estimators interchangeably.
    x = random_state.randn(2, 5, 1).astype(np.float32)
    y_true = x.argsort(axis=1).argsort(axis=1).squeeze(axis=-1)
    return x, y_true


@pytest.mark.parametrize("ranker_name", list(object_rankers.keys()))
def test_object_ranker_fixed(trivial_ranking_problem, ranker_name):
    np.random.seed(123)
    # There are some caveats with pytorch reproducibility. See the comment on
    # the corresponding line of `test_choice_functions.py` for details.
    torch.manual_seed(123)
    torch.use_deterministic_algorithms(True)
    x, y = trivial_ranking_problem
    ranker, params, (loss, acc) = object_rankers[ranker_name]
    ranker = ranker(**params)
    ranker.fit(x, y)
    pred_scores = ranker.predict_scores(x)
    pred_loss = zero_one_rank_loss_for_scores_ties_np(y, pred_scores)
    rtol = 1e-2
    atol = 1e-4
    assert np.isclose(loss, pred_loss, rtol=rtol, atol=atol, equal_nan=False)
    pred = ranker.predict_for_scores(pred_scores)
    pred_2 = ranker.predict(x)
    pred_acc = zero_one_accuracy_np(pred, pred_2)
    assert np.isclose(1.0, pred_acc, rtol=rtol, atol=atol, equal_nan=False)
    pred_acc = zero_one_accuracy_np(pred, y)
    assert np.isclose(acc, pred_acc, rtol=rtol, atol=atol, equal_nan=False)



from .intelligent_system_group_dataset_reader import IntelligentSystemGroupDatasetReader
from .survey_dataset_reader import SurveyDatasetReader

__all__ = [
    "IntelligentSystemGroupDatasetReader",
    "SurveyDatasetReader",
]



import logging

from sklearn.utils import check_random_state

from csrank.constants import DISCRETE_CHOICE
from .util import sub_sampling_discrete_choices_from_relevance
from ..letor_listwise_dataset_reader import LetorListwiseDatasetReader
from ...util import convert_to_label_encoding

logger = logging.getLogger(__name__)


class LetorListwiseDiscreteChoiceDatasetReader(LetorListwiseDatasetReader):
    def __init__(self, random_state=None, n_objects=5, **kwargs):
        super(LetorListwiseDiscreteChoiceDatasetReader, self).__init__(
            learning_problem=DISCRETE_CHOICE, **kwargs
        )
        self.random_state = check_random_state(random_state)
        self.n_objects = n_objects

    def sub_sampling_function(self, X, Y):
        return sub_sampling_discrete_choices_from_relevance(
            Xt=X, Yt=Y, n_objects=self.n_objects
        )

    def convert_output(self, ranking_length):
        self.Y = self.Y.argmin(axis=1)
        self.Y = convert_to_label_encoding(self.Y, ranking_length)

    def splitter(self, iter):
        pass

    def get_train_test_datasets(self, n_datasets):
        pass



"""Apply an evaluation function to instances and collect the results.

The modules that are defined here should take a tensor of instances (shape
:math:`(N, O, F)`, where :math:`N` is the number of instances, :math:`O` the
number of objects per instance and :math:`F` the number of features per object)
and apply some order :math:`k` utility function on all possible combinations of
:math:`k` objects per instance. It should return an evaluation result matrix of
shape :math:`(N, O^k)`.

As an example, consider a pairwise preference function used together with
``PairwiseEvaluation``. If you feed ``PairwiseEvaluation`` with a tensor of
instances, you get back a tensor of preference matrices where the entry at
indes :math:`(i, j)` tells you how preferable object :math:`i` is to object
:math:`j`. You can then combine this result with a reduction to get a
first-order (or order-:math:`k`) utility module. This is used in the
``MeanAggregatedUtility`` first-order utility.
"""

import torch
import torch.nn as nn


def _generate_pairs(instances):
    """Generate object pairs from instances.

    This can be used to generate the individual comparisons in a first-order
    utility function model.

    >>> object_a = [0.5, 0.6]
    >>> object_b = [1.5, 1.6]
    >>> object_c = [2.5, 2.6]
    >>> object_d = [3.5, 3.6]
    >>> object_e = [4.5, 4.6]
    >>> object_f = [5.5, 5.6]
    >>> # instance = list of objects to rank
    >>> instance_a = [object_a, object_b, object_c]
    >>> instance_b = [object_d, object_e, object_f]
    >>> instances = [instance_a, instance_b]

    >>> _generate_pairs(torch.tensor(instances))
    tensor([[[0.5000, 0.6000, 0.5000, 0.6000],
             [0.5000, 0.6000, 1.5000, 1.6000],
             [0.5000, 0.6000, 2.5000, 2.6000],
             [1.5000, 1.6000, 0.5000, 0.6000],
             [1.5000, 1.6000, 1.5000, 1.6000],
             [1.5000, 1.6000, 2.5000, 2.6000],
             [2.5000, 2.6000, 0.5000, 0.6000],
             [2.5000, 2.6000, 1.5000, 1.6000],
             [2.5000, 2.6000, 2.5000, 2.6000]],
    <BLANKLINE>
            [[3.5000, 3.6000, 3.5000, 3.6000],
             [3.5000, 3.6000, 4.5000, 4.6000],
             [3.5000, 3.6000, 5.5000, 5.6000],
             [4.5000, 4.6000, 3.5000, 3.6000],
             [4.5000, 4.6000, 4.5000, 4.6000],
             [4.5000, 4.6000, 5.5000, 5.6000],
             [5.5000, 5.6000, 3.5000, 3.6000],
             [5.5000, 5.6000, 4.5000, 4.6000],
             [5.5000, 5.6000, 5.5000, 5.6000]]])
    """

    def repeat_individual_objects(instances, times):
        """Repeat each object once, immediately after the original."""
        # add a dimension, so that each object is now enclosed in a singleton
        unsqueezed = instances.unsqueeze(2)

        # repeat every object (along the newly added dimension)
        # ([[object_a], [object_a]], [[object_b]], [[object_b]])
        repeated = unsqueezed.repeat(1, 1, times, 1)
        # collapse the added dimension again so that each object is on the same
        # level ([object_a], [object_a], [object_b]], [[object_b])
        return repeated.view(instances.size(0), -1, instances.size(2))

    def repeat_object_lists(instances, times):
        """Repeat the whole object list as a unit (the same as "first" but in a different order)."""
        return instances.repeat(1, times, 1)

    n_objects = instances.size(1)
    first = repeat_individual_objects(instances, n_objects)
    second = repeat_object_lists(instances, n_objects)

    # Glue the two together at the object level (the object's feature vectors
    # are concatenated)
    output_tensor = torch.cat((first, second), dim=2)
    return output_tensor


class PairwiseEvaluation(nn.Module):
    """A module that applies a first-order utility to all object pairs.

    This module receives a list of instances as an input:

    >>> object_00 = [0.5, 0.6]
    >>> object_01 = [1.5, 1.6]
    >>> object_02 = [2.5, 2.6]
    >>> object_10 = [3.5, 3.6]
    >>> object_11 = [4.5, 4.6]
    >>> object_12 = [5.5, 5.6]
    >>> # instance = list of objects to rank
    >>> instance_a = [object_00, object_01, object_02]
    >>> instance_b = [object_10, object_11, object_12]
    >>> instances = torch.tensor([instance_a, instance_b])
    >>> n_features = instances.size(2)

    And computes a list of *relations*, one relation for each instance. A
    relation is a 2d array that contains a score for each pairwise object
    combination.

    >>> import numpy as np
    >>> from csrank.modules.object_mapping import DeterministicSumming
    >>> utility = PairwiseEvaluation(DeterministicSumming(2 * n_features))
    >>> utilities = utility(torch.tensor(instances))
    >>> utilities
    tensor([[[ 2.2000,  4.2000,  6.2000],
             [ 4.2000,  6.2000,  8.2000],
             [ 6.2000,  8.2000, 10.2000]],
    <BLANKLINE>
            [[14.2000, 16.2000, 18.2000],
             [16.2000, 18.2000, 20.2000],
             [18.2000, 20.2000, 22.2000]]])
    >>> utilities[0][0][1] == np.sum(object_00) + np.sum(object_01)
    tensor(True)
    """

    def __init__(self, pairwise_model):
        super().__init__()
        self.pairwise_model = pairwise_model

    def forward(self, instances):
        n_instances = instances.size(0)
        n_objects = instances.size(1)
        object_pairs = _generate_pairs(instances)
        scores = self.pairwise_model(object_pairs)
        return scores.view(n_instances, n_objects, n_objects)



import logging

import numpy as np

from csrank.learner import Learner
from .choice_functions import ChoiceFunctions

logger = logging.getLogger(__name__)


class AllPositive(ChoiceFunctions, Learner):
    def __init__(self, **kwargs):
        """Baseline that assigns the average number of chosen objects in the given choice sets and chooses all the objects

        Parameters
        ----------
        **kwargs: Keyword arguments for the algorithms
        """

    def fit(self, X, Y, **kwd):
        self._pre_fit()
        return self

    def _predict_scores_fixed(self, X, Y, **kwargs):
        return np.zeros_like(Y) + Y.mean()

    def predict_scores(self, X, Y, **kwargs):
        """Predict the utility scores for each object in the collection of set of objects

        Parameters
        ----------
        X : dict or numpy array
           Dictionary with a mapping from ranking size to numpy arrays
           or a single numpy array of size:
           (n_instances, n_objects, n_features)

        Y : dict or numpy array
           Dictionary with a mapping from ranking size to numpy arrays
           or a single numpy array of size:
           (n_instances, n_objects, n_features)
           True scores

        Note that dict and array parameters cannot be mixed.

        Returns
        -------
        Y : dict or numpy array
            Dictionary with a mapping from ranking size to numpy arrays
            or a single numpy array of size:
            (n_instances, n_objects)
            Predicted scores
        """
        logger.info("Predicting scores")

        if (isinstance(X, dict) and not isinstance(Y, dict)) or (
            isinstance(X, np.ndarray) and not isinstance(Y, np.ndarray)
        ):
            raise ValueError(
                "X and Y need to have matching types, either both have to be a dict or both a numpy array."
            )

        if isinstance(X, dict):
            scores = dict()
            for ranking_size, x in X.items():
                scores[ranking_size] = self._predict_scores_fixed(
                    x, Y[ranking_size], **kwargs
                )

        else:
            scores = self._predict_scores_fixed(X, Y, **kwargs)
        return scores

    def predict_for_scores(self, scores, **kwargs):
        if isinstance(scores, dict):
            result = dict()
            for n, score in scores.items():
                result[n] = np.ones_like(score, dtype=int)
        else:
            result = np.ones_like(scores, dtype=int)
        return result



from abc import ABCMeta
import logging

import numpy as np
import skorch
import torch.nn as nn

from csrank.constants import CHOICE_FUNCTION
from csrank.learner import SkorchInstanceEstimator
from csrank.metrics_np import f1_measure
from csrank.util import progress_bar

__all__ = ["ChoiceFunctions"]
logger = logging.getLogger(__name__)


class ChoiceFunctions(metaclass=ABCMeta):
    @property
    def learning_problem(self):
        return CHOICE_FUNCTION

    def predict_for_scores(self, scores, **kwargs):
        """
        Binary choice vector :math:`y` represents the choices amongst the objects in :math:`Q`, such that
        :math:`y(k) = 1` represents that the object :math:`x_k` is chosen and :math:`y(k) = 0` represents it is not
        chosen. Predict choices for the scores for a given collection of sets of objects (query sets).

        Parameters
        ----------
        scores : dict or numpy array
            Dictionary with a mapping from query set size to numpy arrays
            or a single numpy array of size containing scores of each object of size:
            (n_instances, n_objects)


        Returns
        -------
        Y : dict or numpy array
            Dictionary with a mapping from query set size to numpy arrays
            or a single numpy array containing predicted choice vectors of size:
            (n_instances, n_objects)
        """

        if isinstance(scores, dict):
            result = dict()
            for n, score in scores.items():
                result[n] = score > self.threshold_
                result[n] = np.array(result[n], dtype=int)
        else:
            result = scores > self.threshold_
            result = np.array(result, dtype=int)
        return result

    def _tune_threshold(self, X_val, Y_val, thin_thresholds=1, verbose=0):
        scores = self.predict_scores(X_val)
        probabilities = np.unique(scores)[::thin_thresholds]
        threshold = 0.0
        best = f1_measure(Y_val, scores > threshold)
        try:
            for i, p in enumerate(probabilities):
                pred = scores > p
                f1 = f1_measure(Y_val, pred)
                if f1 > best:
                    threshold = p
                    best = f1
                if verbose == 1:
                    progress_bar(i, len(probabilities), status="Tuning threshold")
        except KeyboardInterrupt:
            logger.info("Keyboard interrupted")
        logger.info(
            "Tuned threshold, obtained {:.2f} which achieved"
            " a micro F1-measure of {:.2f}".format(threshold, best)
        )
        return threshold


class SkorchChoiceFunction(ChoiceFunctions, SkorchInstanceEstimator):
    """A variable choice estimator based on some scoring module.

    This estimator takes a scoring module and combines it with a sigmoid
    activation to predict scores between 0 and 1. The choice is then made based
    on a fixed threshold value. This makes it very simple to derive new
    estimators with any given scoring function. Refer to skorch's documentation
    for supported parameters. For example the optimizer or the optimizer's
    learning rate could be overridden.

    Parameters
    ----------
    module : torch module (class)
        This is the scoring module. It should be an uninstantiated
        ``torch.nn.Module`` class that expects the number of features per
        object as its only parameter on initialization.

    criterion : torch criterion (class)
        The criterion that is used to evaluate and optimize the module.

    threshold : float
        The threshold value that is used to convert scores to a choice. Must be
        between 0 and 1. Defaults to 0.5.

    **kwargs : skorch NeuralNet arguments
        All keyword arguments are passed to the constructor of
        ``skorch.NeuralNet``. See the documentation of that class for more
        details.
    """

    def __init__(self, module, criterion=nn.BCELoss, threshold=0.5, **kwargs):
        super().__init__(module=module, criterion=criterion, **kwargs)
        # The scoring is trained to predict something close to "0" for
        # non-chosen values, something close to "1" for chosen values. So 0.5
        # is a natural threshold. It would be possible to additionally tune
        # that threshold.
        self.threshold_ = threshold

    def initialize_module(self, *args, **kwargs):
        params = self.get_params_for("module")
        # Add a Sigmoid activation since the resulting "scores" should be
        # between 0 and 1.
        self.module_ = nn.Sequential(self.module(**params), nn.Sigmoid())
        self.module_ = skorch.utils.to_device(self.module_, self.device)



import numpy as np
from sklearn.utils import check_random_state

from csrank.constants import CHOICE_FUNCTION
from ..synthetic_dataset_generator import SyntheticDatasetGenerator


class ChoiceDatasetGenerator(SyntheticDatasetGenerator):
    def __init__(self, dataset_type="pareto", **kwargs):
        super(ChoiceDatasetGenerator, self).__init__(
            learning_problem=CHOICE_FUNCTION, **kwargs
        )
        dataset_function_options = {
            "linear": self.make_latent_linear_choices,
            "pareto": self.make_globular_pareto_choices,
        }
        if dataset_type not in dataset_function_options:
            raise ValueError(
                f"dataset_type must be one of {set(dataset_function_options.keys())}"
            )
        self.dataset_function = dataset_function_options[dataset_type]

    def make_globular_pareto_choices(
        self,
        n_instances=10000,
        n_features=2,
        n_objects=10,
        seed=42,
        cluster_spread=1.0,
        **kwargs,
    ):
        def pareto_front(X, signs=None):
            n_points, n_attributes = X.shape
            if signs is None:
                signs = -np.ones(n_attributes)
            pareto = np.ones(n_points, dtype=bool)
            for i, attr in enumerate(X):
                pareto[i] = np.all(
                    np.any((X * signs[None, :]) <= (attr * signs), axis=1)
                )
            return pareto

        def sample_from_unit_ball(n_points, dimension, radius, random_state):
            """Sample points uniformly from a ball.

            The ball has radius `radius` and is centered at the origin.

            Parameters
            ----------
            n_points : int
                The number of points to sample.
            dimension : int
                The dimension of the space.
            radius : float
                The radius of the ball.
            random_state: np.random.RandomState
                A numpy random state.

            Returns
            -------
            numpy array of shape (n_points, dimension)
                A list of points sampled from the ball.
            """
            # Sample a random direction for each point
            directions = random_state.randn(n_points, dimension)
            # Normalize each direction vector to have length 1 (euclidean
            # norm).
            directions /= np.linalg.norm(directions, axis=1, ord=2)[:, None]

            # Sample a length (as a fraction of the radius) uniformly for each
            # point.
            u = random_state.uniform(size=n_points)[:, None]
            lengths = u * radius

            return directions * lengths

        def sample_pareto_from_isometric_normal(
            n_points, dimension, center, random_state
        ):
            """Generate a Pareto problem from random objects.

            Objects are drawn from a d-dimensional isometric normal
            distribution.

            This should be the easiest possible Pareto-problem, since the model can learn
            a latent-utility which scores how likely a point is on the front (independent
            of the other points).

            Parameters
            ----------
            n_points : int
                The number of points to sample.
            dimension : int
                The dimension of the space.
            center : scalar or numpy array
                An offset that will be added to every point.
            random_state: np.random.RandomState
                A numpy random state.

            Returns
            -------
            X: numpy array of shape (n_points, dimension)
                A list of points sampled from the d-dimensional isometric
                normal distribution.
            Y. numpy array of shape n_points
                A binary flag array indicating whether or not the corresponding
                point is part of the Pareto front.
            """
            X = random_state.randn(n_points, dimension)
            Y = pareto_front(X)
            return X + center, Y

        rand = check_random_state(seed)
        X = np.empty((n_instances, n_objects, n_features))
        Y = np.empty((n_instances, n_objects), dtype=int)
        for i in range(n_instances):
            center = sample_from_unit_ball(
                n_points=1,
                dimension=n_features,
                radius=cluster_spread,
                random_state=rand,
            )
            x, y = sample_pareto_from_isometric_normal(
                n_points=n_objects,
                dimension=n_features,
                center=center,
                random_state=rand,
            )
            X[i] = x
            Y[i] = y
        return X, Y

    def make_latent_linear_choices(
        self,
        n_instances=10000,
        n_features=2,
        n_objects=6,
        n_rep_units=5,
        threshold=0.0,
        seed=42,
        **kwargs,
    ):
        rand = check_random_state(seed)
        ranw = check_random_state(rand.randint(2 ** 32, dtype="uint32"))
        X = rand.uniform(-1, 1, size=(n_instances, n_objects, n_features))
        W_rep = ranw.randn(n_features, n_rep_units)
        rep = X.dot(W_rep).mean(axis=-2)
        w_join = ranw.randn(n_features + n_rep_units)
        joint_matrix = np.empty(
            (n_objects, n_instances, n_features + n_rep_units), dtype=np.float32
        )
        for i in range(n_objects):
            joint_matrix[i] = np.concatenate((X[:, i], rep), axis=-1)
        scores = joint_matrix.dot(w_join)
        Y = scores > threshold
        Y = Y.astype(int)
        return X, Y.T



import logging
import os

import numpy as np
from sklearn.model_selection import train_test_split
from sklearn.utils import check_random_state

from csrank.dataset_reader.dataset_reader import DatasetReader
from csrank.dataset_reader.util import standardize_features

logger = logging.getLogger(__name__)


class MNISTDatasetReader(DatasetReader):
    def __init__(
        self,
        n_train_instances=10000,
        n_test_instances=10000,
        n_objects=10,
        random_state=None,
        standardize=True,
        **kwargs,
    ):
        super(MNISTDatasetReader, self).__init__(dataset_folder="mnist", **kwargs)
        self.n_test_instances = n_test_instances
        self.n_train_instances = n_train_instances
        self.n_objects = n_objects
        self.random_state = check_random_state(random_state)
        self.n_features = None
        self.standardize = standardize
        self.__load_dataset__()
        logger.info("Done loading the dataset")

    def __load_dataset__(self):
        x_file = os.path.join(self.dirname, "X_raw_features.npy")
        y_file = os.path.join(self.dirname, "y_labels.npy")
        self.X_raw = np.load(x_file)
        self.y_labels = np.load(y_file)
        self.n_features = self.X_raw.shape[1]

    def get_single_train_test_split(self):
        self.dataset_function()
        x_train, x_test, y_train, y_test = train_test_split(
            self.X,
            self.Y,
            random_state=self.random_state,
            test_size=self.n_test_instances,
        )
        if self.standardize:
            x_train, x_test = standardize_features(x_train, x_test)
        logger.info("Done")
        return x_train, y_train, x_test, y_test

    def splitter(self, iter):
        pass

    def get_dataset_dictionaries(self, lengths=[5, 6]):
        x_train = dict()
        y_train = dict()
        x_test = dict()
        y_test = dict()
        for n_obj in lengths:
            self.n_objects = n_obj
            self.dataset_function()
            x_1, x_2, y_1, y_2 = train_test_split(
                self.X,
                self.Y,
                random_state=self.random_state,
                test_size=self.n_test_instances,
            )
            if self.standardize:
                x_1, x_2 = standardize_features(x_1, x_2)
            x_train[n_obj], x_test[n_obj], y_train[n_obj], y_test[n_obj] = (
                x_1,
                x_2,
                y_1,
                y_2,
            )
        logger.info("Done")
        return x_train, y_train, x_test, y_test

    def get_train_test_datasets(self, n_datasets=5):
        pass
